#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.field_type_methods import *
from reclaimer.h2.constants import *


def h2_rawdata_ref_parser(self, desc, node=None, parent=None, attr_index=None,
                          rawdata=None, root_offset=0, offset=0, **kwargs):
    try:
        orig_offset = offset
        if node is None:
            parent[attr_index] = node = desc.get(NODE_CLS, self.node_cls)\
                (desc, parent=parent, init_attrs=rawdata is None)

        # If there is rawdata to build the structure from
        if rawdata is not None:
            offsets = desc[ATTR_OFFS]
            # loop once for each field in the node
            for i in range(len(node)):
                desc[i][TYPE].parser(desc[i], None, node, i, rawdata,
                                       root_offset, offset + offsets[i],
                                       **kwargs)

            # increment offset by the size of the struct
            offset += desc[SIZE]

        s_desc = desc.get(STEPTREE)
        if s_desc:
            pointer_converter = kwargs.get("map_pointer_converter")
            if pointer_converter is not None:
                file_ptr = pointer_converter.v_ptr_to_f_ptr(node[1])
                if not node[0] or (file_ptr + node[0] > len(rawdata) or
                                   file_ptr <= 0):
                    # data is stored in a resource map, or the size is invalid
                    s_desc[TYPE].parser(s_desc, None, node, STEPTREE, None)
                else:
                    s_desc[TYPE].parser(
                        s_desc, None, node, STEPTREE, rawdata,
                        root_offset, file_ptr, **kwargs)
            elif 'steptree_parents' not in kwargs:
                offset = s_desc['TYPE'].parser(s_desc, None, node, 'STEPTREE',
                                               rawdata, root_offset, offset,
                                               **kwargs)
            else:
                kwargs['steptree_parents'].append(node)

        # pass the incremented offset to the caller
        return offset
    except Exception as e:
        # if the error occurred while parsing something that doesnt have an
        # error report routine built into the function, do it for it.
        kwargs.update(buffer=rawdata, root_offset=root_offset)
        if 's_desc' in locals():
            e = format_parse_error(e, field_type=s_desc.get(TYPE), desc=s_desc,
                                  parent=node, attr_index=STEPTREE,
                                  offset=offset, **kwargs)
        elif 'i' in locals():
            e = format_parse_error(e, field_type=desc[i].get(TYPE),
                                   desc=desc[i], parent=node, attr_index=i,
                                   offset=offset, **kwargs)
        e = format_parse_error(e, field_type=self, desc=desc,
                               parent=parent, attr_index=attr_index,
                               offset=orig_offset, **kwargs)
        raise e


def h2_tag_ref_parser(self, desc, node=None, parent=None, attr_index=None,
                      rawdata=None, root_offset=0, offset=0, **kwargs):
    # parse tag_class and id(if meta) or path length(if not meta)
    try:
        orig_offset = offset
        if node is None:
            parent[attr_index] = node = desc.get(NODE_CLS, self.node_cls)\
                (desc, parent=parent, init_attrs=rawdata is None)

        # If there is rawdata to build the structure from
        if rawdata is not None:
            offsets = desc[ATTR_OFFS]
            if "map_pointer_converter" in kwargs:
                attr_indices = (0, 1)
                node[2] = 0
            else:
                attr_indices = (0, 2)
                node[1] = 0xFFffFFff

            # loop once for each field in the node
            for i in attr_indices:
                desc[i][TYPE].parser(desc[i], None, node, i, rawdata,
                                     root_offset, offset + offsets[i],
                                     **kwargs)

            # increment offset by the size of the struct
            offset += desc[SIZE]

        if 'steptree_parents' in kwargs:
            kwargs['steptree_parents'].append(node)
        else:
            s_desc = desc.get(STEPTREE)
            offset = s_desc[TYPE].parser(s_desc, None, node, STEPTREE,
                                         rawdata, root_offset, offset,
                                         **kwargs)

        # pass the incremented offset to the caller
        return offset
    except Exception as e:
        # if the error occurred while parsing something that doesnt have an
        # error report routine built into the function, do it for it.
        kwargs.update(buffer=rawdata, root_offset=root_offset)
        if 's_desc' in locals():
            e = format_parse_error(e, field_type=s_desc.get(TYPE), desc=s_desc,
                                  parent=node, attr_index=STEPTREE,
                                  offset=offset, **kwargs)
        elif 'i' in locals():
            e = format_parse_error(e, field_type=desc[i].get(TYPE),
                                   desc=desc[i], parent=node, attr_index=i,
                                   offset=offset, **kwargs)
        e = format_parse_error(e, field_type=self, desc=desc,
                               parent=parent, attr_index=attr_index,
                               offset=orig_offset, **kwargs)
        raise e



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: xXTxMANXx
# 	Mapped Color Amounts.
# revision: 3		author: Lord Zedd
# 	Standardizing.
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


ltvl_light_volume_system_unknown_0 = Struct("unknown_0",
    SInt16("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=2
    )


ltvl_light_volume_system_import_data_function = Struct("function",
    SInt32("unknown_0", VISIBLE=False),
    h3_string_id("name"),
    BytesRaw("unknown_1", SIZE=8, VISIBLE=False),
    h3_rawdata_ref("function"),
    ENDIAN=">", SIZE=36
    )


ltvl_light_volume_system_import_data = Struct("import_data",
    h3_string_id("material_type"),
    SInt32("unknown_0", VISIBLE=False),
    h3_dependency("bitmap"),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    SInt32("unknown_2", VISIBLE=False),
    Array("unknown_array", SUB_STRUCT=SInt16("unknown"), SIZE=6, VISIBLE=False),
    BytesRaw("unknown_3", SIZE=4, VISIBLE=False),
    h3_reflexive("functions", ltvl_light_volume_system_import_data_function),
    ENDIAN=">", SIZE=60
    )


ltvl_light_volume_system_shader_propertie_shader_map = Struct("shader_map",
    h3_dependency("bitmap"),
    SInt8("unknown_0", VISIBLE=False),
    SInt8("bitmap_index"),
    SInt8("unknown_1", VISIBLE=False),
    Bool8("bitmap_flags", *unknown_flags_8),
    SInt8("unknown_bitmap_index_enable"),
    SInt8("uv_argument_index"),
    SInt8("unknown_2", VISIBLE=False),
    SInt8("unknown_3", VISIBLE=False),
    ENDIAN=">", SIZE=24
    )


ltvl_light_volume_system_shader_propertie_argument = Struct("argument",
    Array("arg_array", SUB_STRUCT=Float("arg"), SIZE=4),
    ENDIAN=">", SIZE=16
    )


ltvl_light_volume_system_shader_propertie_unknown_0 = Struct("unknown_0",
    BytesRaw("unknown", SIZE=4, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


ltvl_light_volume_system_shader_propertie_unknown_2 = Struct("unknown_2",
    SInt16("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=2
    )


ltvl_light_volume_system_shader_propertie_unknown_3 = Struct("unknown_3",
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    SInt8("unknown_1", VISIBLE=False),
    SInt8("unknown_2", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=6
    )


ltvl_light_volume_system_shader_propertie_unknown_4 = Struct("unknown_4",
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=4
    )


ltvl_light_volume_system_shader_propertie_function = Struct("function",
    SInt32("unknown_0", VISIBLE=False),
    h3_string_id("name"),
    BytesRaw("unknown_1", SIZE=8, VISIBLE=False),
    h3_rawdata_ref("function"),
    ENDIAN=">", SIZE=36
    )


ltvl_light_volume_system_shader_propertie = Struct("shader_propertie",
    h3_dependency("template"),
    h3_reflexive("shader_maps", ltvl_light_volume_system_shader_propertie_shader_map),
    h3_reflexive("arguments", ltvl_light_volume_system_shader_propertie_argument),
    h3_reflexive("unknown_0", ltvl_light_volume_system_shader_propertie_unknown_0),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    h3_reflexive("unknown_2", ltvl_light_volume_system_shader_propertie_unknown_2),
    h3_reflexive("unknown_3", ltvl_light_volume_system_shader_propertie_unknown_3),
    h3_reflexive("unknown_4", ltvl_light_volume_system_shader_propertie_unknown_4),
    h3_reflexive("functions", ltvl_light_volume_system_shader_propertie_function),
    SInt32("unknown_5", VISIBLE=False),
    SInt32("unknown_6", VISIBLE=False),
    BytesRaw("unknown_7", SIZE=4, VISIBLE=False),
    Array("unknown_array", SUB_STRUCT=SInt16("unknown"), SIZE=8, VISIBLE=False),
    ENDIAN=">", SIZE=132
    )


ltvl_light_volume_system_unknown_20 = Struct("unknown_20",
    BytesRaw("unknown", SIZE=16, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


ltvl_light_volume_system_compiled_function = Struct("compiled_function",
    BytesRaw("unknown", SIZE=64, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=64
    )


ltvl_light_volume_system_compiled_color_function = Struct("compiled_color_function",
    color_rgb_float("color"),
    Float("magnitude"),
    ENDIAN=">", SIZE=16
    )


ltvl_light_volume_system = Struct("light_volume_system",
    h3_string_id("name"),
    h3_dependency("base_render_method"),
    h3_reflexive("unknown_0", ltvl_light_volume_system_unknown_0),
    h3_reflexive("import_data", ltvl_light_volume_system_import_data),
    h3_reflexive("shader_properties", ltvl_light_volume_system_shader_propertie),
    Array("unknown_array", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    BytesRaw("unknown_1", SIZE=4, VISIBLE=False),
    SInt32("unknown_2", VISIBLE=False),
    BytesRaw("unknown_3", SIZE=8, VISIBLE=False),
    SInt8("input_0"),
    SInt8("input_range_0"),
    SEnum8("output_kind_0", *cntl_contrail_system_output_kind_0),
    SInt8("output_0"),
    h3_rawdata_ref("unknown_4"),
    BytesRaw("unknown_5", SIZE=8, VISIBLE=False),
    SInt8("input_1"),
    SInt8("input_range_1"),
    SEnum8("output_kind_1", *cntl_contrail_system_output_kind_0),
    SInt8("output_1"),
    h3_rawdata_ref("unknown_6"),
    BytesRaw("unknown_7", SIZE=8, VISIBLE=False),
    SInt8("input_2"),
    SInt8("input_range_2"),
    SEnum8("output_kind_2", *cntl_contrail_system_output_kind_0),
    SInt8("output_2"),
    h3_rawdata_ref("unknown_8"),
    BytesRaw("unknown_9", SIZE=8, VISIBLE=False),
    SInt8("input_3"),
    SInt8("input_range_3"),
    SEnum8("output_kind_3", *cntl_contrail_system_output_kind_0),
    SInt8("output_3"),
    h3_rawdata_ref("unknown_10"),
    BytesRaw("unknown_11", SIZE=8, VISIBLE=False),
    SInt8("input_4"),
    SInt8("input_range_4"),
    SEnum8("output_kind_4", *cntl_contrail_system_output_kind_0),
    SInt8("output_4"),
    h3_rawdata_ref("unknown_12"),
    BytesRaw("unknown_13", SIZE=8, VISIBLE=False),
    SInt8("input_5"),
    SInt8("input_range_5"),
    SEnum8("output_kind_5", *cntl_contrail_system_output_kind_0),
    SInt8("output_5"),
    h3_rawdata_ref("unknown_14"),
    BytesRaw("unknown_15", SIZE=8, VISIBLE=False),
    SInt8("input_6"),
    SInt8("input_range_6"),
    SEnum8("output_kind_6", *cntl_contrail_system_output_kind_0),
    SInt8("output_6"),
    h3_rawdata_ref("unknown_16"),
    BytesRaw("unknown_17", SIZE=8, VISIBLE=False),
    SInt8("input_7"),
    SInt8("input_range_7"),
    SEnum8("output_kind_7", *cntl_contrail_system_output_kind_0),
    SInt8("output_7"),
    h3_rawdata_ref("unknown_18"),
    BytesRaw("unknown_19", SIZE=20, VISIBLE=False),
    h3_reflexive("unknown_20", ltvl_light_volume_system_unknown_20),
    h3_reflexive("compiled_functions", ltvl_light_volume_system_compiled_function),
    h3_reflexive("compiled_color_functions", ltvl_light_volume_system_compiled_color_function),
    ENDIAN=">", SIZE=380
    )


ltvl_body = Struct("tagdata",
    h3_reflexive("light_volume_system", ltvl_light_volume_system),
    ENDIAN=">", SIZE=12
    )


def get():
    return ltvl_def

ltvl_def = TagDef("ltvl",
    h3_blam_header('ltvl'),
    ltvl_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["ltvl"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.senv import *
from .shdr import *
from supyr_struct.defs.tag_def import TagDef

senv_body = Struct("tagdata",
    shdr_attrs,
    senv_attrs,
    SIZE=836,
    )

def get():
    return senv_def

senv_def = TagDef("senv",
    blam_header_stubbs('senv', 2),
    senv_body,

    ext=".shader_environment", endian=">", tag_cls=ShdrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from supyr_struct.defs.tag_def import TagDef
from ..common_descs import *

unknown_float_struct = Struct("unknown_float_struct",
    # f0 and f1 sum to 1.0
    Struct("floats", Float("f0"), Float("f1"), ORIENT='h'),
    Struct("color", INCLUDE=rgb_float),
    )

stage_map = Struct("stage_map",
    Pad(32),
    SInt32("unknown0"),
    Struct("unknown_animation", INCLUDE=anim_src_func_per_pha_sca),

    Struct("unknown1", Float("f0"), Float("f1"), ORIENT='h'),
    dependency_stubbs("bitmap", "bitm"),

    Pad(32),
    Struct("u_animation", INCLUDE=anim_src_func_per_pha_sca),
    Struct("v_animation", INCLUDE=anim_src_func_per_pha_sca),
    Struct("rotation_animation", INCLUDE=anim_src_func_per_pha_sca_rot),
    SIZE=164
    )

floats_9 = tuple(Float("unknown" + str(i)) for i in range(9))
floats_7 = tuple(Float("unknown" + str(i)) for i in range(7))

imef_body = Struct("tagdata",
    Struct("unknown_floats0", *floats_9),
    Struct("color0", INCLUDE=rgb_float),
    Struct("color1", INCLUDE=rgb_float),
    Struct("unknown_animation0", INCLUDE=anim_src_func_per_pha_sca),
    Pad(4),
    Struct("unknown_struct1", INCLUDE=unknown_float_struct),
    Struct("unknown_floats1", *floats_9),
    Struct("color2", INCLUDE=rgb_float),
    Struct("unknown_struct2", INCLUDE=unknown_float_struct),
    Struct("unknown_floats2", *floats_7),
    SInt16("unknown0"),
    Struct("unknown_animation1", INCLUDE=anim_src_func_per_pha_sca),
    Struct("unknown_floats3", *floats_7),

    Pad(156),
    reflexive("stage_maps", stage_map, OFFSET=416),
    SIZE=428
    )

def get():
    return imef_def

imef_def = TagDef("imef",
    blam_header_os('imef', 1),
    imef_body,

    ext=".image_effect", endian=">"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: DeadCanadian
# 	named the tags and Verticies
# revision: 3		author: Lord Zedd
# 	port h2
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


ant__vertice = Struct("vertice",
    yp_float_rad("angle"),
    Float("length"),
    SInt16("sequence_index"),
    SInt16("unknown", VISIBLE=False),
    color_argb_float("color"),
    color_argb_float("lod_color"),
    Pad(16),
    ENDIAN=">", SIZE=64
    )


ant__body = Struct("tagdata",
    h3_string_id("attachment_marker_name"),
    h3_dependency("bitmaps"),
    h3_dependency("physics"),
    BytesRaw("unknown", SIZE=28, VISIBLE=False),
    h3_reflexive("vertices", ant__vertice),
    ENDIAN=">", SIZE=76
    )


def get():
    return ant__def

ant__def = TagDef("ant!",
    h3_blam_header('ant!'),
    ant__body,

    ext=".%s" % h3_tag_class_fcc_to_ext["ant!"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.devi import DeviTag
from supyr_struct.defs.tag_def import TagDef

devi_attrs = Struct("devi_attrs",
    Bool32("flags",
        "position_loops",
        "position_not_interpolated",
        ),

    float_sec("power_transition_time"),
    float_sec("power_acceleration_time"),
    float_sec("position_transition_time"),
    float_sec("position_acceleration_time"),
    float_sec("depowered_position_transition_time"),
    float_sec("depowered_position_acceleration_time"),

    SEnum16("A_in", *device_functions),
    SEnum16("B_in", *device_functions),
    SEnum16("C_in", *device_functions),
    SEnum16("D_in", *device_functions),

    dependency("open", valid_event_effects),
    dependency("close", valid_event_effects),
    dependency("opened", valid_event_effects),
    dependency("closed", valid_event_effects),
    dependency("depowered", valid_event_effects),
    dependency("repowered", valid_event_effects),

    float_sec("delay_time"),
    Pad(8),
    dependency("delay_effect", valid_event_effects),
    float_wu("automatic_activation_radius"),

    Pad(84),
    FlFloat("inv_power_acceleration_time"),
    FlFloat("inv_power_transition_time"),
    FlFloat("inv_depowered_acceleration_time"),
    FlFloat("inv_depowered_transition_time"),
    FlFloat("inv_position_acceleration_time"),
    FlFloat("inv_position_transition_time"),
    Pad(4),

    SIZE=276,
    )

devi_body = Struct('tagdata',
    devi_attrs,
    SIZE=276
    )

def get():
    return devi_def

devi_def = TagDef("devi",
    blam_header('devi'),
    devi_body,

    ext=".device", endian=">", tag_cls=DeviTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from ...hek.defs.objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

child_ids = Struct("child_id",
    dyn_senum32("entry_reference",
        DYN_NAME_PATH="tagdata.entries.STEPTREE[DYN_I].name.data"),
    SIZE=4
    )

referenced_by = Struct("referenced_by", INCLUDE=child_ids)

entry = Struct("entry",
    rawtext_ref("name", StrUtf8, max_size=256, widget=EntryFrame),
    UEnum32("group_tag", INCLUDE=valid_tags_os),
    SInt32("handle_data"),
    SInt32("flags"),
    reflexive("child_ids", child_ids, 65536),
    reflexive("referenced_by", referenced_by, 65536),
    SIZE=68
    )

tag__body = Struct("tagdata",
    reflexive("entries", entry, 65536, DYN_NAME_PATH='.name.data'),
    SIZE=36
    )

def get():
    return tag__def

tag__def = TagDef("tag+",
    blam_header_os('tag+'),
    tag__body,

    ext=".tag_database", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = ( 'MergedJmsModel', )


from .model import JmsModel
from .node import JmsNode
from .material import JmsMaterial
from .merged_region import MergedJmsRegion


class MergedJmsModel:
    node_list_checksum = 0
    nodes = ()
    materials = ()
    regions = ()

    u_scale = 1.0
    v_scale = 1.0

    def __init__(self, *jms_models):
        self.nodes = []
        self.materials = []
        self.regions = {}

        for jms_model in jms_models:
            self.merge_jms_model(jms_model)

    def calc_uv_scales(self):
        u_scale = self.u_scale
        v_scale = self.v_scale
        calc_u_scale = 0.0
        calc_v_scale = 0.0
        for region in self.regions.values():
            for perm_mesh in region.perm_meshes.values():
                for meshes in perm_mesh.lod_meshes.values():
                    for lod_mesh_list in perm_mesh.lod_meshes.values():
                        for mesh in lod_mesh_list.values():
                            for vert in mesh.verts:
                                calc_u_scale = max(abs(vert.tex_u * u_scale),
                                                   calc_u_scale)
                                calc_v_scale = max(abs(vert.tex_v * v_scale),
                                                   calc_v_scale)

        return calc_u_scale, calc_v_scale

    verify_models_match = JmsModel.verify_models_match

    def merge_jms_model(self, other_model):
        all_errors = {}
        first_nodes = None

        if not other_model:
            return

        if not self.nodes:
            self.node_list_checksum = other_model.node_list_checksum
            self.nodes = []
            self.materials = []
            for node in other_model.nodes:
                self.nodes.append(
                    JmsNode(
                        node.name, node.first_child, node.sibling_index,
                        node.rot_i, node.rot_j, node.rot_k, node.rot_w,
                        node.pos_x, node.pos_y, node.pos_z, node.parent_index)
                    )

            for mat in other_model.materials:
                self.materials.append(
                    JmsMaterial(
                        mat.name, mat.tiff_path, mat.shader_path,
                        mat.shader_type, mat.properties)
                    )

            self.regions = {}

        errors = self.verify_models_match(other_model)
        if errors:
            return errors

        new_mat_counts = {}
        for mat in self.materials:
            new_mat_counts[mat.name] = new_mat_counts.get(mat.name, 0) - 1

        default_mats = {}
        for mat in other_model.materials:
            default_mats.setdefault(mat.name, mat)
            new_mat_counts[mat.name] = new_mat_counts.get(mat.name, 0) + 1

        for mat_name, mat_ct in new_mat_counts.items():
            if mat_ct > 0:
                self.materials.extend((default_mats[mat_name], ) * mat_ct)

        # merge each region from the other model into this ones regions
        for region in other_model.regions:
            if region not in self.regions:
                self.regions[region] = MergedJmsRegion(region)

            self.regions[region].merge_jms_model(other_model, self.materials)

        # correct the region index numbers for each marker in the regions
        i = 0
        for region_name in sorted(self.regions):
            perm_meshes = self.regions[region_name].perm_meshes
            for perm_mesh in perm_meshes.values():
                for marker in perm_mesh.markers:
                    marker.region = i
            i += 1

        return all_errors



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .obje import *
from .item import *
from .objs.weap import WeapTag

# replace the object_type enum one that uses
# the correct default value for this object
obje_attrs = dict(obje_attrs)
obje_attrs[0] = dict(obje_attrs[0], DEFAULT=2)

magazine_item = Struct("magazine_item",
    SInt16("rounds"),

    Pad(10),
    dependency('equipment', "eqip"),
    SIZE=28
    )

magazine = Struct("magazine",
    Bool32("flags",
        "wastes_rounds_when_reloaded",
        "every_round_must_be_chambered"
        ),
    SInt16("rounds_recharged", SIDETIP="per second"),
    SInt16("rounds_total_initial"),
    SInt16("rounds_total_maximum"),
    SInt16("rounds_loaded_maximum"),

    Pad(8),
    float_sec("reload_time"),
    SInt16("rounds_reloaded"),

    Pad(2),
    float_sec("chamber_time"),

    Pad(24),
    dependency('reloading_effect', valid_event_effects),
    dependency('chambering_effect', valid_event_effects),

    Pad(12),
    reflexive("magazine_items", magazine_item, 2,
        "primary", "secondary"),
    SIZE=112
    )

firing_effect = Struct("firing_effect",
    SInt16("shot_count_lower_bound"),
    SInt16("shot_count_upper_bound"),

    Pad(32),
    dependency('firing_effect', valid_event_effects),
    dependency('misfire_effect', valid_event_effects),
    dependency('empty_effect', valid_event_effects),
    dependency('firing_damage', "jpt!"),
    dependency('misfire_damage', "jpt!"),
    dependency('empty_damage', "jpt!"),
    SIZE=132
    )

trigger = Struct("trigger",
    Bool32("flags",
        "tracks_fired_projectile",
        "random_firing_effects",
        "can_fire_with_partial_ammo",
        "does_not_repeat_automatically",
        "locks_in_on_off_state",
        "projectiles_use_weapon_origin",
        "sticks_when_dropped",
        "ejects_during_chamber",
        "discharging_spews",
        "analog_rate_of_fire",
        "use_error_when_unzoomed",
        "projectile_vector_cannot_be_adjusted",
        "projectiles_have_identical_error",
        "projectile_is_client_side_only",
        ),
    Struct("firing",
        QStruct("rounds_per_second",
            Float("from", UNIT_SCALE=per_sec_unit_scale, GUI_NAME=''),
            Float("to",   UNIT_SCALE=per_sec_unit_scale),
            ORIENT='h'
            ),
        float_sec("acceleration_time"),
        float_sec("deceleration_time"),
        Float("blurred_rate_of_fire", UNIT_SCALE=per_sec_unit_scale),

        Pad(8),
        SEnum16("magazine",
            'primary',
            'secondary',
            ('NONE', -1),
            DEFAULT=-1
            ),
        SInt16("rounds_per_shot"),
        SInt16("minimum_rounds_loaded"),
        SInt16("rounds_between_tracers"),

        Pad(6),
        SEnum16("firing_noise", *sound_volumes),
        from_to_zero_to_one("error"),
        float_sec("error_acceleration_time"),
        float_sec("error_deceleration_time"),
        ),

    Pad(8),
    Struct("charging",
        float_sec("charging_time"),
        float_sec("charge_hold_time"),
        SEnum16("overcharged_action",
            'none',
            'explode',
            'discharge',
            ),

        Pad(2),
        float_zero_to_one("charged_illumination"),
        float_sec("spew_time"),
        dependency('charging_effect', valid_event_effects),
        ),

    Struct("projectile",
        SEnum16("distribution_function",
            'point',
            'horizontal_fan',
            ),
        SInt16("projectiles_per_shot"),
        float_deg("distribution_angle"),  # degrees

        Pad(4),
        float_rad("minimum_error"),  # radians
        from_to_rad("error_angle"),  # radians
        QStruct("first_person_offset", INCLUDE=xyz_float),

        Pad(4),
        dependency('projectile', valid_objects),
        ),

    Struct("misc",
        float_sec("ejection_port_recovery_time"),
        float_sec("illumination_recovery_time"),

        Pad(12),
        float_zero_to_one("heat_generated_per_round"),
        float_zero_to_one("age_generated_per_round"),

        Pad(4),
        float_sec("overload_time"),

        Pad(40),
        ),

    QStruct("misc_rates",
        Float("ejection_port_recovery_rate"),
        Float("illumination_recovery_rate"),
        Float("acceleration_rate"),
        Float("deceleration_rate"),
        Float("error_acceleration_rate"),
        Float("error_deceleration_rate"),
        VISIBLE=False,
        COMMENT="\
These are various rates that are calculated when the weapon is compiled into a map."
        ),
    reflexive("firing_effects", firing_effect, 8),

    SIZE=276
    )

weap_attrs = Struct("weap_attrs",
    Bool32("flags",
        "vertical_heat_display",
        "mutually_exclusive_triggers",
        "attacks_automatically_on_bump",
        "must_be_readied",
        "doesnt_count_toward_maximum",
        "aim_assists_only_when_zoomed",
        "prevents_grenade_throwing",
        "must_be_picked_up",
        "holds_triggers_when_dropped",
        "prevents_melee_attack",
        "detonates_when_dropped",
        "cannot_fire_at_maximum_age",
        "secondary_trigger_overrides_grenades",
        "does_not_depower_active_camo_in_multiplayer",  # obsolete
        "enables_integrated_night_vision",
        "ai_uses_weapon_melee_damage"
        ),
    ascii_str32('label'),
    SEnum16('secondary_trigger_mode',
        "normal",
        "slaved_to_primary",
        "inhibits_primary",
        "loads_alternate_ammunition",
        "loads_multiple_primary_ammunition",
        ),
    SInt16("max_alternate_shots_loaded"),
    SEnum16('A_in', *weapon_export_to),
    SEnum16('B_in', *weapon_export_to),
    SEnum16('C_in', *weapon_export_to),
    SEnum16('D_in', *weapon_export_to),
    float_sec("ready_time"),
    dependency('ready_effect', valid_event_effects),

    Struct("heat",
        float_zero_to_one("recovery_threshold"),
        float_zero_to_one("overheated_threshold"),
        float_zero_to_one("detonation_threshold"),
        float_zero_to_one("detonation_fraction"),
        float_zero_to_inf("loss_per_second", UNIT_SCALE=per_sec_unit_scale),
        float_zero_to_one("illumination"),

        Pad(16),
        dependency('overheated', valid_event_effects),
        dependency('detonation', valid_event_effects),
        ),

    Struct("melee",
        dependency('player_damage', "jpt!"),
        dependency('player_response', "jpt!"),
        ),

    Pad(8),
    Struct("aiming",
        dependency("actor_firing_parameters", "actv"),
        float_wu("near_reticle_range"),  # world units
        float_wu("far_reticle_range"),  # world units
        float_wu("intersection_reticle_range"),  # world units

        Pad(2),
        SInt16("zoom_levels"),
        QStruct("zoom_ranges", INCLUDE=from_to),
        float_rad("autoaim_angle"),  # radians
        float_wu("autoaim_range"),  # world units
        float_rad("magnetism_angle"),  # radians
        float_wu("magnetism_range"),  # world units
        float_rad("deviation_angle"),  # radians
        ),

    Pad(4),
    Struct("movement",
        SEnum16('penalized',
            "always",
            "when_zoomed",
            "when_zoomed_or_reloading",
            ),
        Pad(2),
        Float("forward_penalty"),
        Float("sideways_penalty"),
        ),

    Pad(4),
    Struct("ai_targeting",
        Float("minimum_target_range"),
        Float("looking_time_modifier")
        ),

    Pad(4),
    Struct("light",
        float_sec('power_on_time',  UNIT_SCALE=sec_unit_scale),
        float_sec('power_off_time', UNIT_SCALE=sec_unit_scale),
        dependency('power_on_effect', valid_event_effects),
        dependency('power_off_effect', valid_event_effects)
        ),

    Struct("age",
        Float("heat_penalty"),
        Float("rate_of_fire_penalty"),
        float_zero_to_one("misfire_start"),
        float_zero_to_one("misfire_chance")
        ),

    Pad(12),
    Struct("interface",
        dependency('first_person_model', valid_models),
        dependency('first_person_animations', "antr"),

        Pad(4),
        dependency('hud_interface', "wphi"),
        dependency('pickup_sound', "snd!"),
        dependency('zoom_in_sound', "snd!"),
        dependency('zoom_out_sound', "snd!"),

        Pad(12),
        Float('active_camo_ding'),
        Float('active_camo_regrowth_rate', UNIT_SCALE=per_sec_unit_scale),
        ),

    Pad(14),
    SEnum16('weapon_type', *weapon_types),

    reflexive("predicted_resources", predicted_resource, 1024, VISIBLE=False),
    reflexive("magazines", magazine, 2,
        "primary", "secondary"),
    reflexive("triggers", trigger, 2,
        "primary", "secondary"),

    SIZE=512
    )

weap_body = Struct("tagdata",
    obje_attrs,
    item_attrs,
    weap_attrs,
    SIZE=1288,
    )


def get():
    return weap_def

weap_def = TagDef("weap",
    blam_header('weap', 2),
    weap_body,

    ext=".weapon", endian=">", tag_cls=WeapTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: OrangeMohawk
# 	Rumble/Camera Shake from RMBL/CSDT
# revision: 3		author: Lord Zedd
# 	Updating and standardizing
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


drdf_response = Struct("response",
    SEnum16("response_type", *jpt__player_response_response_type),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SInt16("unknown_2", VISIBLE=False),
    BytesRaw("unknown_3", SIZE=12, VISIBLE=False),
    Float("unknown_4", VISIBLE=False),
    BytesRaw("unknown_5", SIZE=8, VISIBLE=False),
    Float("unknown_6", VISIBLE=False),
    Float("unknown_7", VISIBLE=False),
    SInt16("unknown_8", VISIBLE=False),
    SInt16("unknown_9", VISIBLE=False),
    Array("unknown_array_0", SUB_STRUCT=Float("unknown"), SIZE=4, VISIBLE=False),
    SInt16("unknown_10", VISIBLE=False),
    SInt16("unknown_11", VISIBLE=False),
    Array("unknown_array_1", SUB_STRUCT=Float("unknown"), SIZE=7, VISIBLE=False),
    Float("low_frequency_vibration_duration"),
    h3_rawdata_ref("low_frequency_vibration_function"),
    Float("high_frequency_vibration_duration"),
    h3_rawdata_ref("high_frequency_vibration_function"),
    Float("duration_0"),
    SEnum16("fade_function", *jpt__player_response_fade_function),
    SInt16("unknown_12", VISIBLE=False),
    float_rad("rotation"),
    Float("pushback"),
    QStruct("jitter", INCLUDE=from_to),
    Float("duration_1"),
    SEnum16("falloff_function", *jpt__player_response_fade_function),
    SInt16("unknown_13", VISIBLE=False),
    Float("random_translation"),
    float_rad("random_rotation"),
    SEnum16("wobble_function", *jpt__wobble_function),
    SInt16("unknown_14", VISIBLE=False),
    Float("wobble_function_period"),
    Float("wobble_weight"),
    ENDIAN=">", SIZE=192
    )


drdf_body = Struct("tagdata",
    h3_reflexive("responses", drdf_response),
    ENDIAN=">", SIZE=12
    )


def get():
    return drdf_def

drdf_def = TagDef("drdf",
    h3_blam_header('drdf'),
    drdf_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["drdf"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from copy import copy, deepcopy

try:
    from mozzarilla.widgets.field_widgets import ReflexiveFrame,\
         HaloRawdataFrame, TextFrame, ColorPickerFrame, EntryFrame,\
         SoundSampleFrame, DynamicArrayFrame, Halo3BitmapTagFrame
except Exception:
    ReflexiveFrame = HaloRawdataFrame = TextFrame = ColorPickerFrame =\
                     EntryFrame = SoundSampleFrame = DynamicArrayFrame = \
                     Halo3BitmapTagFrame = None
from reclaimer.common_descs import *
from reclaimer.h3.field_types import H3TagRef, H3Reflexive, H3RawdataRef
from reclaimer.h3.constants import DYN_NAME_PATH, STEPTREE, NAME_MAP, WIDGET,\
     TOOLTIP, COMMENT, MAX, DEFAULT, VISIBLE, SIZE, MAX_REFLEXIVE_COUNT,\
     h3_tag_class_fcc_to_ext
from reclaimer.h3.enums import *

from supyr_struct.defs.block_def import BlockDef


def h3_tag_class(*args, **kwargs):
    '''
    A macro for creating a tag_class enum desc with the
    enumerations set to the provided tag_class fcc's.
    '''
    classes = []
    for four_cc in args:
        classes.append((h3_tag_class_fcc_to_ext[four_cc], four_cc))

    return UEnum32(
        'tag_class',
        *(tuple(sorted(classes)) + (("NONE", 0xffffffff),) ),
        DEFAULT=0xffffffff, GUI_NAME='', WIDGET_WIDTH=20, **kwargs
        )

def h3_reflexive(name, substruct, max_count=MAX_REFLEXIVE_COUNT, *names, **desc):
    '''This function serves to macro the creation of a reflexive'''
    desc.update(
        INCLUDE=h3_reflexive_struct,
        STEPTREE=Array(name + " array",
            SIZE=".size", MAX=max_count,
            SUB_STRUCT=substruct, WIDGET=ReflexiveFrame
            ),
        SIZE=12
        )
    if DYN_NAME_PATH in desc:
        desc[STEPTREE][DYN_NAME_PATH] = desc.pop(DYN_NAME_PATH)
    if names:
        name_map = {}
        for i in range(len(names)):
            e_name = BlockDef.str_to_name(None, names[i])
            name_map[e_name] = i

        desc[STEPTREE][NAME_MAP] = name_map

    return H3Reflexive(name, **desc)


def h3_rawdata_ref(name, f_type=BytearrayRaw, max_size=None,
                   widget=HaloRawdataFrame, **kwargs):
    '''This function serves to macro the creation of a rawdata reference'''
    ref_struct = dict(h3_rawdata_ref_struct)
    if COMMENT in kwargs: ref_struct[COMMENT] = kwargs.pop(COMMENT)
    if TOOLTIP in kwargs: ref_struct[TOOLTIP] = kwargs.pop(TOOLTIP)
    if max_size is not None:
        ref_struct[0] = dict(ref_struct[0])
        ref_struct[0][MAX] = max_size
        kwargs[MAX] = max_size

    if widget is not None:
        kwargs[WIDGET] = widget

    return H3RawdataRef(name,
        INCLUDE=ref_struct,
        STEPTREE=f_type("data", GUI_NAME="", SIZE=".size", **kwargs))


def h3_dependency(name='tag ref', valid_ids=None, **kwargs):
    '''This function serves to macro the creation of a tag dependency'''
    if isinstance(valid_ids, tuple):
        valid_ids = h3_tag_class(*valid_ids)
    elif isinstance(valid_ids, str):
        valid_ids = h3_tag_class(valid_ids)
    elif valid_ids is None:
        valid_ids = valid_h3_tags

    return H3TagRef(name,
        valid_ids,
        INCLUDE=tag_ref_struct,
        STEPTREE=StrTagRef(
            "filepath", SIZE=tag_ref_str_size, GUI_NAME="", MAX=254),
        **kwargs
        )

def h3_blam_header(tagid, version=1):
    '''This function serves to macro the creation of a tag header'''
    header_desc = dict(h3_tag_header)
    header_desc[1] = dict(header_desc[1])
    header_desc[5] = dict(header_desc[5])
    header_desc[1][DEFAULT] = tagid
    header_desc[5][DEFAULT] = version
    return header_desc


valid_h3_tags = h3_tag_class(*h3_tag_class_fcc_to_ext.keys())


# ###########################################################################
# The order of element in all the enumerators is important(DONT SHUFFLE THEM)
# ###########################################################################

#Shared Enumerator options
old_materials_list = materials_list
# old materials list that is referenced in tags, but seems to be unused
del materials_list


# Descriptors
h3_blam_engine_id = dict(blam_engine_id)
h3_blam_engine_id[DEFAULT] = "b3am"
h3_tag_header = Struct("blam_header",
    Pad(36),
    UEnum32("tag class",
        GUI_NAME="tag class", INCLUDE=valid_h3_tags, EDITABLE=False
        ),
    UInt32("checksum", DEFAULT=0x4D6F7A7A, EDITABLE=False),
    UInt32("header size", DEFAULT=64, EDITABLE=False),
    Bool64("flags",
        "edited with mozz",
        EDITABLE=False
        ),
    UInt16("version", DEFAULT=1, EDITABLE=False),
    UInt8("integrity0", DEFAULT=0, EDITABLE=False),
    UInt8("integrity1", DEFAULT=255, EDITABLE=False),
    h3_blam_engine_id,
    VISIBLE=VISIBILITY_METADATA, SIZE=64
    )

# these structs replace the above ones when used outside a map
h3_tag_index_ref_struct = H3TagRef('dependency',
    valid_h3_tags,
    SInt32("path pointer", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    SInt32("path length", MAX=254, VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    UInt32("id", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    ORIENT='h'
    )

h3_reflexive_struct = H3Reflexive('reflexive',
    SInt32("size", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    UInt32("pointer", VISIBLE=VISIBILITY_METADATA, EDITABLE=False, DEFAULT=0xFFFFFFFF),
    UInt32("id", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    )

h3_rawdata_ref_struct = H3RawdataRef('rawdata ref',
    SInt32("size", GUI_NAME="", SIDETIP="bytes", EDITABLE=False),
    Bool32("flags", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    UInt32("raw pointer", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    UInt32("pointer", VISIBLE=VISIBILITY_METADATA, EDITABLE=False, DEFAULT=0xFFFFFFFF),
    UInt32("id", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    ORIENT='h'
    )

h3_page_offset_info_struct = BitStruct("page_offset_info",
    UBitInt("offset", SIZE=29),
    Bit("fixup"),
    Bit("raw_page"),
    )


def h3_string_id(name, **kwargs):
    return string_id(name, 16, 8, 8, **kwargs)


def h3_page_offset_info(name):
    return BitStruct(name, INCLUDE=h3_page_offset_info_struct)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .obje import *
from .devi import *
from .objs.ctrl import CtrlTag
from supyr_struct.defs.tag_def import TagDef

# replace the object_type enum one that uses
# the correct default value for this object
obje_attrs = dict(obje_attrs)
obje_attrs[0] = dict(obje_attrs[0], DEFAULT=8)

ctrl_attrs = Struct("ctrl_attrs",
    SEnum16('type',
        'toggle_switch',
        'on_button',
        'off_button',
        'call_button'
        ),
    SEnum16('triggers_when',
        'touched_by_player',
        'destroyed'
        ),
    float_zero_to_one('call_value'),

    Pad(80),
    dependency("on", valid_event_effects),
    dependency("off", valid_event_effects),
    dependency("deny", valid_event_effects),
    )

ctrl_body = Struct("tagdata",
    obje_attrs,
    devi_attrs,
    ctrl_attrs,

    SIZE=792,
    )


def get():
    return ctrl_def

ctrl_def = TagDef("ctrl",
    blam_header('ctrl'),
    ctrl_body,

    ext=".device_control", endian=">", tag_cls=CtrlTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import math
import os
import re
import traceback

from copy import deepcopy
from pathlib import Path
from reclaimer.common_descs import anim_types, anim_frame_info_types
from reclaimer.model.jms import JmsNode, JmsModel
from reclaimer.util import float_to_str, float_to_str_truncate,\
     parse_jm_float, parse_jm_int
from reclaimer.util.matrices import clip_angle_to_bounds, are_vectors_equal,\
     quat_to_matrix, matrix_to_quat, axis_angle_to_quat, quat_to_axis_angle,\
     quaternion_to_euler, euler_to_quaternion, Quaternion, multiply_quaternions


JMA_ANIMATION_EXTENSIONS = (
    ".jma", ".jmm", ".jmo", ".jmr", ".jmt", ".jmw", ".jmz",
    )


def get_anim_ext(anim_type, frame_info_type, world_relative=False):
    anim_type = anim_type.lower()
    frame_info_type = frame_info_type.lower()
    if anim_type == "replacement":
        return ".jmr"
    elif anim_type == "overlay":
        return ".jmo"
    elif "dz" in frame_info_type:
        return ".jmz"
    elif "dyaw" in frame_info_type:
        return ".jmt"
    elif "dx" in frame_info_type and "dy" in frame_info_type:
        return ".jma"
    elif world_relative:
        return ".jmw"
    else:
        return ".jmm"


def get_anim_types(anim_ext):
    anim_ext = anim_ext.lower()
    if "jmr" in anim_ext:
        return anim_types[2], anim_frame_info_types[0], False
    elif "jmo" in anim_ext:
        return anim_types[1], anim_frame_info_types[0], False
    elif "jmz" in anim_ext:
        return anim_types[0], anim_frame_info_types[3], False
    elif "jmt" in anim_ext:
        return anim_types[0], anim_frame_info_types[2], False
    elif "jma" in anim_ext:
        return anim_types[0], anim_frame_info_types[1], False
    else:
        return anim_types[0], anim_frame_info_types[0], "jmw" in anim_ext


class JmaRootNodeState:
    __slots__ = (
        "dx", "dy", "dz", "dyaw",
        "x", "y", "z", "yaw",
        )
    def __init__(self, dx=0.0, dy=0.0, dz=0.0, dyaw=0.0,
                 x=0.0, y=0.0, z=0.0, yaw=0.0):
        self.dx, self.dy, self.dz, self.dyaw = dx, dy, dz, dyaw
        self.x,  self.y,  self.z,  self.yaw  =  x,  y,  z,  yaw

    def __repr__(self):
        return """JmaRootNodeState(
    dx=%s, dy=%s, dz=%s, dyaw=%s,
    x=%s, y=%s, z=%s, yaw=%s,
)""" % (self.dx, self.dy, self.dz, self.dyaw,
        self.x, self.y, self.z, self.yaw)

    def __eq__(self, other):
        if not isinstance(other, JmaRootNodeState):
            return False
        elif (abs(self.dx - other.dx) > 0.0000000001 or
              abs(self.dy - other.dy) > 0.0000000001 or
              abs(self.dz - other.dz) > 0.0000000001 or
              abs(self.dyaw - other.dyaw) > 0.0000000001):
            return False
        return True


class JmaNodeState:
    __slots__ = (
        "pos_x", "pos_y", "pos_z",
        "rot_i", "rot_j", "rot_k", "rot_w",
        "scale",
        )
    def __init__(self, pos_x=0.0, pos_y=0.0, pos_z=0.0,
                 rot_i=0.0, rot_j=0.0, rot_k=0.0, rot_w=1.0, scale=1.0):
        self.rot_i = rot_i
        self.rot_j = rot_j
        self.rot_k = rot_k
        self.rot_w = rot_w
        self.pos_x = pos_x
        self.pos_y = pos_y
        self.pos_z = pos_z
        self.scale = scale

    def __repr__(self):
        return """JmaNodeState(
    x=%s, y=%s, z=%s,
    i=%s, j=%s, k=%s, w=%s,
    scale=%s
)""" % (self.pos_x, self.pos_y, self.pos_z,
        self.rot_i, self.rot_j, self.rot_k, self.rot_w,
        self.scale)

    def __eq__(self, other):
        if not isinstance(other, JmaNodeState):
            return False
        elif (abs(self.rot_i - other.rot_i) > 0.000001 or
              abs(self.rot_j - other.rot_j) > 0.000001 or
              abs(self.rot_k - other.rot_k) > 0.000001 or
              abs(self.rot_w - other.rot_w) > 0.000001):
            return False
        elif (abs(self.pos_x - other.pos_x) > 0.000001 or
              abs(self.pos_y - other.pos_y) > 0.000001 or
              abs(self.pos_z - other.pos_z) > 0.000001):
            return False
        elif abs(self.scale - other.scale) > 0.000001:
            return False
        return True

    def __sub__(self, other):
        if not isinstance(other, JmaNodeState):
            raise TypeError("Cannot subtract %s from %s" %
                            (type(other), type(self)))
        return JmaNodeState(
            self.pos_x - other.pos_x, self.pos_y - other.pos_y,
            self.pos_z - other.pos_z, self.rot_i - other.rot_i,
            self.rot_j - other.rot_j, self.rot_k - other.rot_k,
            self.rot_w - other.rot_w, self.scale - other.scale,
            )


class JmaAnimation:
    name = ""
    node_list_checksum = 0
    nodes = ()
    frames = ()

    rot_keyframes = ()
    trans_keyframes = ()
    scale_keyframes = ()

    root_node_info = ()

    frame_rate = 30
    actors = ()

    world_relative = False

    rot_flags_int   = 0
    trans_flags_int = 0
    scale_flags_int = 0

    root_node_info_applied = True

    anim_type = anim_types[0]
    frame_info_type = anim_frame_info_types[0]

    def __init__(self, name="", node_list_checksum=0,
                 anim_type="", frame_info_type="", world_relative=False,
                 nodes=None, frames=None, actors=None, frame_rate=30):

        self.name = name.strip(" ")
        self.node_list_checksum = node_list_checksum
        self.nodes  = nodes  if nodes  else []
        self.frames = frames if frames else []
        self.world_relative = bool(world_relative)
        self.anim_type = anim_type
        self.frame_info_type = frame_info_type
        self.frame_rate = frame_rate
        self.actors = actors if actors else ["unnamedActor"]

        self.root_node_info = []
        self.setup_keyframes()

    def setup_keyframes(self):
        self.rot_keyframes   = [[] for i in range(self.node_count)]
        self.trans_keyframes = [[] for i in range(self.node_count)]
        self.scale_keyframes = [[] for i in range(self.node_count)]

    @property
    def has_keyframe_data(self):
        if not self.rot_keyframes or len(self.rot_keyframes) != self.node_count:
            return False
        elif not self.trans_keyframes or len(self.trans_keyframes) != self.node_count:
            return False
        elif not self.scale_keyframes or len(self.scale_keyframes) != self.node_count:
            return False
        return True

    @property
    def ext(self):
        return get_anim_ext(self.anim_type, self.frame_info_type,
                            self.world_relative)
    @property
    def has_dxdy(self): return "dx"   in self.frame_info_type
    @property
    def has_dyaw(self): return "dyaw" in self.frame_info_type
    @property
    def has_dz(self):   return "dz"   in self.frame_info_type

    @property
    def actor_count(self): return len(self.actors)
    @property
    def frame_count(self): return len(self.frames)
    @property
    def node_count(self): return len(self.nodes)

    @property
    def last_frame_loops_to_first(self):
        return self.anim_type != "overlay"

    @property
    def trans_flags(self):
        return [bool(self.trans_flags_int & (1 << i))
                for i in range(self.node_count)]

    @property
    def rot_flags(self):
        return [bool(self.rot_flags_int & (1 << i))
                for i in range(self.node_count)]

    @property
    def scale_flags(self):
        return [bool(self.scale_flags_int & (1 << i))
                for i in range(self.node_count)]

    @property
    def root_node_info_frame_size(self):
        if self.has_dz:
            return 16
        elif self.has_dyaw:
            return 12
        elif self.has_dxdy:
            return 8
        return 0

    @property
    def frame_data_frame_size(self):
        return (12 * sum(self.trans_flags) +
                8  * sum(self.rot_flags) +
                4  * sum(self.scale_flags))

    @property
    def default_data_size(self):
        return self.node_count * 24 - self.frame_data_frame_size

    def get_node_index(self, node_name):
        node_name = node_name.lower()
        i = 0
        for node in self.nodes:
            if node.name.lower() == node_name:
                return i
            i += 1
        return -1

    def get_root_node_info(self):
        return list(self.root_node_info)

    def get_node(self, node_name_or_index):
        if isinstance(node_name_or_index):
            node_name_or_index = self.get_node_index(node_name_or_index)

        if node_name_or_index in range(self.node_count):
            return self.nodes[node_name_or_index]
    def get_node_frames(self, node_name_or_index):
        if isinstance(node_name_or_index):
            node_name_or_index = self.get_node_index(node_name_or_index)

        if node_name_or_index in range(self.node_count):
            return [frame[node_name_or_index] for frame in self.frames]

    def set_frame(self, frame_index, frame_data=None, frame_info=None):
        assert frame_index in range(-(self.frame_count + 1), self.frame_count + 1)

        if frame_data is not None:
            self.frames[frame_index] = frame_data
        if frame_info is not None:
            self.root_node_info[frame_index] = frame_info
    def set_node_frames(self, node_index, frame_data):
        assert node_index in range(-(self.node_count + 1), self.node_count + 1)
        for f in range(self.frame_count):
            self.frames[f][node_index] = frame_data[f]

    def append_frame(self, frame_data=None, frame_info=None):
        self.insert_frame(self.frame_count, frame_data, frame_info)
    def append_frames(self, frames):
        for frame in frames:
            self.insert_frame(self.frame_count, *frame)

    def append_node(self, node, frame_data=None):
        self.insert_node(self.node_count, node, frame_data)
    def append_nodes(self, nodes_and_frames):
        for node_and_frames in nodes_and_frames:
            assert len(nodes_and_frames) > 0
            self.insert_node(self.node_count, *node_and_frames)

    def insert_frame(self, frame_index, frame_data=None, frame_info=None):
        assert frame_index in range(-(self.frame_count + 1), self.frame_count + 1)

        if frame_data is None:
            frame_data = [JmaNodeState() for i in range(self.node_count)]
        else:
            assert len(frame_data) == self.frame_count
            for node_frame in frame_data:
                assert isinstance(node_frame, JmaNodeState)

        if frame_info is None:
            frame_info = JmaRootNodeState()
        else:
            assert isinstance(frame_info, JmaRootNodeState)

        self.frames.insert(frame_index, frame_data)
        self.root_node_info.insert(frame_index, frame_info)
    def insert_frames(self, frames_by_indices):
        for frame_index in reversed(sorted(frames_by_indices)):
            self.insert_frame(frame_index, *frames_by_indices[frame_index])

    def insert_node(self, node_index, node, frame_data=None):
        assert isinstance(node, JmsNode)
        if frame_data is None:
            frame_data = [JmaNodeState() for i in range(self.frame_count)]
        else:
            assert len(frame_data) == self.frame_count
            for node_frame in frame_data:
                assert isinstance(node_frame, JmaNodeState)

        for f in range(self.frame_count):
            self.frames[f].insert(node_index, frame_data[f])
        self.nodes.insert(node_index, node)
    def insert_nodes(self, nodes_and_frames_by_indices):
        for node_index in reversed(sorted(nodes_and_frames_by_indices)):
            self.insert_node(node_index, *nodes_and_frames_by_indices[node_index])

    def remove_frame(self, frame_index):
        assert frame_index in range(-self.frame_count, self.frame_count)
        if frame_index in range(len(self.root_node_info)):
            return (self.frames.pop(frame_index),
                    self.root_node_info.pop(frame_index))
        return self.frames.pop(frame_index), None
    def remove_frames(self, frame_indices):
        return {int(i): self.remove_frame(i) for i in
                reversed(sorted(frame_indices))}

    def remove_node(self, node_index):
        assert node_index in range(-self.node_count, self.node_count)
        return (self.nodes.pop(node_index),
                [frame.pop(node_index) for frame in self.frames])
    def remove_nodes(self, node_indices):
        return {int(i): self.remove_node(i) for i in
                reversed(sorted(node_indices))}

    def apply_root_node_info_to_states(self, undo=False):
        if self.root_node_info_applied == (not undo):
            # do nothing if the root node info is already applied
            # and we are being told to apply it, or its not applied
            # and we are being told to undo its application.
            return

        if self.has_dxdy or self.has_dz or self.has_dyaw:
            delta = -1 if undo else 1
            for f in range(self.frame_count):
                # apply the total change in the root nodes
                # frame_info for this frame to the frame_data
                node_info = self.root_node_info[f]
                node_state = self.frames[f][0]

                q0 = Quaternion(euler_to_quaternion(
                    0, 0, -node_info.yaw * delta)).normalized
                q1 = Quaternion((node_state.rot_i, node_state.rot_j,
                                 node_state.rot_k, node_state.rot_w)).normalized

                i, j, k, w = multiply_quaternions(q0, q1)

                node_state.pos_x += node_info.x * delta
                node_state.pos_y += node_info.y * delta
                node_state.pos_z += node_info.z * delta
                node_state.rot_i = i
                node_state.rot_j = j
                node_state.rot_k = k
                node_state.rot_w = w

        self.root_node_info_applied = not undo

    def calculate_root_node_info(self):
        self.root_node_info = []

        has_dxdy, has_dz, has_dyaw = self.has_dxdy, self.has_dz, self.has_dyaw

        dx = dy = dz = dyaw = x = y = z = yaw = 0.0
        frame_count = self.frame_count
        for f in range(frame_count):
            node_state0 = self.frames[f][0]
            node_state1 = self.frames[(f + 1) % frame_count][0]
            if has_dxdy:
                dx = node_state1.pos_x - node_state0.pos_x
                dy = node_state1.pos_y - node_state0.pos_y

            if has_dz:
                dz = node_state1.pos_z - node_state0.pos_z

            if has_dyaw:
                q0 = (node_state0.rot_i, node_state0.rot_j,
                      node_state0.rot_k, node_state0.rot_w)
                q1 = (node_state1.rot_i, node_state1.rot_j,
                                node_state1.rot_k, node_state1.rot_w)

                # remove the rotation of the next node from this one to
                # find the yaw difference between the 2 quaternions
                ex, ey, ez = quaternion_to_euler(
                    *multiply_quaternions(q0, Quaternion(q1).inverse)
                    )
                dyaw = clip_angle_to_bounds(ez)

            if f + 1 == frame_count:
                dx = dy = dz = dyaw = 0.0

            self.root_node_info.append(
                JmaRootNodeState(dx, dy, dz, dyaw, x, y, z, yaw)
                )
            x += dx
            y += dy
            z += dz
            yaw += dyaw

    verify_nodes_valid = JmsModel.verify_nodes_valid
    get_node_depths = JmsModel.get_node_depths

    def calculate_animation_flags(self, tolerance=None):
        if tolerance is None:
            tolerance = 0

        self.rot_flags_int = 0
        self.trans_flags_int = 0
        self.scale_flags_int = 0
        if len(self.frames) < 2:
            return

        f0 = self.frames[0]

        zero = (0, 0, 0, 0)
        # determine which transforms types of each node are animated
        # by seeing how much they change from the starting frame
        for node_states in self.frames[1: ]:
            for n in range(len(node_states)):
                diff = node_states[n] - f0[n]
                if not are_vectors_equal(
                        (diff.rot_i, diff.rot_j, diff.rot_k, diff.rot_w),
                        zero, False, tolerance):
                    self.rot_flags_int |= (1 << n)

                if not are_vectors_equal(
                        (diff.pos_x, diff.pos_y, diff.pos_z),
                        zero, False, tolerance):
                    self.trans_flags_int |= (1 << n)

        # scale is calculated a bit differently. we ALWAYS store scale
        # frame data for a node if its scale is ever not 1.0, even if
        # it is static for the entire animation length. This seems to
        # be due to how compressed animations handle default scales.
        for node_states in self.frames:
            for n in range(len(node_states)):
                if not are_vectors_equal(
                        (node_states[n].scale, ),
                        (1.0, ), False, tolerance):
                    self.scale_flags_int |= (1 << n)

    def verify_animations_match(self, other_jma):
        errors = list(other_jma.verify_jma())
        if len(other_jma.nodes) != len(self.nodes):
            errors.append("Node counts do not match.")
            return errors

        for i in range(len(self.nodes)):
            if not self.nodes[i].is_node_hierarchy_equal(other_jma.nodes[i]):
                errors.append("Nodes '%s' do not match." % i)

        return errors

    def verify_jma(self):
        crc = self.node_list_checksum
        node_error = False
        node_ct = len(self.nodes)

        errors = self.verify_nodes_valid()
        if errors:
            return errors

        for frame in self.frames:
            if len(frame) != node_ct:
                errors.append("Invalid node state count for frame(s).")
                break

        return errors

    def compress_animation(self, **kw):
        raise NotImplementedError("Compression isn't implemented yet.")


class JmaAnimationSet:
    node_list_checksum = 0
    nodes = ()
    animations = ()

    def __init__(self, *jma_animations):
        self.nodes = []
        self.animations = {}

        for jma_animation in jma_animations:
            self.merge_jma_animation(jma_animation)

    verify_animations_match = JmaAnimation.verify_animations_match

    def merge_jma_animation(self, other_jma):
        assert isinstance(other_jma, JmaAnimation)

        if not other_jma:
            return

        if not self.nodes:
            self.node_list_checksum = other_jma.node_list_checksum
            self.nodes = []
            for node in other_jma.nodes:
                self.nodes.append(
                    JmsNode(
                        node.name, node.first_child, node.sibling_index,
                        node.rot_i, node.rot_j, node.rot_k, node.rot_w,
                        node.pos_x, node.pos_y, node.pos_z, node.parent_index)
                    )

        errors = self.verify_animations_match(other_jma)
        if errors:
            return errors

        self.animations[other_jma.name] = other_jma

        return errors


def read_jma(jma_string, stop_at="", anim_name=""):
    if anim_name is None:
        anim_name = "__unnamed"

    anim_name, ext = os.path.splitext(anim_name)
    anim_type, frame_info_type, world_relative = get_anim_types(ext)

    jma_anim = JmaAnimation(anim_name, 0, anim_type,
                            frame_info_type, world_relative)
    jma_string = jma_string.replace("\n", "\t")

    data = tuple(d for d in jma_string.split("\t") if d)
    dat_i = 0

    if parse_jm_int(data[dat_i]) != 16392:
        print("JMA identifier '16392' not found.")
        return jma_anim

    dat_i += 1

    try:
        frame_count = parse_jm_int(data[dat_i]) & 0xFFffFFff
        dat_i += 1
    except Exception:
        print(traceback.format_exc())
        print("Could not read frame count.")
        return jma_anim

    if frame_count > 2048:
        raise ValueError("Cannot parse jma files with more than 2048 frames.")

    try:
        frame_rate = parse_jm_int(data[dat_i]) & 0xFFffFFff
        dat_i += 1
    except Exception:
        print(traceback.format_exc())
        print("Could not frame rate.")
        return jma_anim

    if stop_at == "actors": return jma_anim

    try:
        jma_anim.actors = [None] * (parse_jm_int(data[dat_i]) & 0xFFffFFff)
        dat_i += 1
    except Exception:
        print(traceback.format_exc())
        print("Failed to read actor count.")
        del jma_anim.actors[:]
        return jma_anim

    if jma_anim.actor_count != 1:
        raise ValueError("Cannot parse jma files with more than one actor.")

    # read the actors
    for actor_i in range(len(jma_anim.actors)):
        # TODO: update getting these if multiple actors are ever supported
        nodes  = jma_anim.nodes
        frames = jma_anim.frames
        try:
            jma_anim.actors[actor_i] = data[dat_i]
            dat_i += 1
        except Exception:
            print(traceback.format_exc())
            print("Failed to read actors.")
            del jma_anim.actors[actor_i: ]
            return jma_anim

        try:
            node_count = parse_jm_int(data[dat_i]) & 0xFFffFFff
            dat_i += 1
        except Exception:
            print(traceback.format_exc())
            print("Could not node count.")
            return jma_anim

        if node_count > 256:
            raise ValueError("Cannot parse jma files with more than 256 nodes.")

        if stop_at == "checksum": continue

        try:
            jma_anim.node_list_checksum = parse_jm_int(data[dat_i])
            dat_i += 1
        except Exception:
            print(traceback.format_exc())
            print("Could not read node list checksum.")
            return jma_anim

        if stop_at == "nodes": continue

        # read the nodes
        try:
            i = 0  # make sure i is defined in case of exception
            nodes[:] = [None] * node_count
            for i in range(node_count):
                nodes[i] = JmsNode(
                    data[dat_i], parse_jm_int(data[dat_i+1]), parse_jm_int(data[dat_i+2])
                    )
                dat_i += 3
            JmsNode.setup_node_hierarchy(nodes)
        except Exception:
            print(traceback.format_exc())
            print("Failed to read nodes.")
            del nodes[i: ]
            return jma_anim


        if stop_at == "frames": continue

        # read the frame data
        try:
            i = 0 # make sure i is defined in case of exception
            for i in range(frame_count):
                frame = [None] * node_count
                for j in range(node_count):
                    frame[j] = JmaNodeState(
                        parse_jm_float(data[dat_i]),   parse_jm_float(data[dat_i+1]),
                        parse_jm_float(data[dat_i+2]), parse_jm_float(data[dat_i+3]),
                        parse_jm_float(data[dat_i+4]), parse_jm_float(data[dat_i+5]),
                        parse_jm_float(data[dat_i+6]), parse_jm_float(data[dat_i+7])
                        )
                    dat_i += 8
                frames.append(frame)
        except Exception:
            print(traceback.format_exc())
            print("Failed to read frames.")
            del frames[i: ]
            return jma_anim

    jma_anim.calculate_root_node_info()
    jma_anim.apply_root_node_info_to_states(True)
    jma_anim.calculate_animation_flags()
    return jma_anim


def write_jma(filepath, jma_anim, use_blitzkrieg_rounding=False):
    if jma_anim.actor_count != 1:
        raise ValueError("Cannot write jma files with more than one actor.")

    if use_blitzkrieg_rounding:
        to_str = lambda f: float_to_str_truncate(f, 6)
    else:
        to_str = float_to_str

    # If the path doesnt exist, create it
    filepath = Path(filepath)
    filepath.parent.mkdir(exist_ok=True, parents=True)

    if not jma_anim.root_node_info_applied:
        jma_anim = deepcopy(jma_anim)
        jma_anim.apply_root_node_info_to_states()

    with filepath.open("w", encoding='latin1', newline="\r\n") as f:
        f.write("16392\n")  # version number
        f.write("%s\n" % jma_anim.frame_count)
        f.write("%s\n" % jma_anim.frame_rate)

        f.write("%s\n" % jma_anim.actor_count)

        for actor_name in jma_anim.actors:
            # TODO: update getting these if multiple actors are ever supported
            nodes  = jma_anim.nodes
            frames = jma_anim.frames

            f.write("%s\n" % actor_name)
            f.write("%s\n" % len(nodes))
            f.write("%s\n" % int(jma_anim.node_list_checksum))

            for node in nodes:
                f.write("%s\n%s\n%s\n" %
                    (node.name[: 31], node.first_child, node.sibling_index)
                )

            for frame in frames:
                for nf in frame:
                    f.write("%s\t%s\t%s\n%s\t%s\t%s\t%s\n%s\n" % (
                        to_str(nf.pos_x), to_str(nf.pos_y), to_str(nf.pos_z),
                        to_str(nf.rot_i), to_str(nf.rot_j),
                        to_str(nf.rot_k), to_str(nf.rot_w),
                        to_str(nf.scale))
                    )

#jma_file = open(r'C:\Users\Moses\Desktop\halo\data\characters\cyborg\animations\alert unarmed turn-right.jmt', 'r')
#data = read_jma(jma_file.read(), "", "alert unarmed turn-right.jmt")



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

def get(): return udlg_def

def snd_dependency(name):
    return dependency(name, "snd!")

udlg_body = Struct("tagdata",
    Pad(16),
    Struct("idle",
        snd_dependency("noncombat"),
        snd_dependency("combat"),
        snd_dependency("flee"),
        Pad(48),
        ),

    Struct("involuntary",
        snd_dependency("pain_body_minor"),
        snd_dependency("pain_body_maior"),
        snd_dependency("pain_shield"),
        snd_dependency("pain_falling"),
        snd_dependency("scream_fear"),
        snd_dependency("scream_pain"),
        snd_dependency("maimed_limb"),
        snd_dependency("maimed_head"),
        snd_dependency("death_quiet"),
        snd_dependency("death_violent"),
        snd_dependency("death_falling"),
        snd_dependency("death_agonizing"),
        snd_dependency("death_instant"),
        snd_dependency("death_flying"),
        Pad(16),
        ),

    Struct("hurting_people",
        snd_dependency("damaged_friend"),
        snd_dependency("damaged_friend_player"),
        snd_dependency("damaged_enemy"),
        snd_dependency("damaged_enemy_cm"),
        Pad(64),
        ),

    Struct("being_hurt",
        snd_dependency("hurt_friend"),
        snd_dependency("hurt_friend_re"),
        snd_dependency("hurt_friend_player"),
        snd_dependency("hurt_enemy"),
        snd_dependency("hurt_enemy_re"),
        snd_dependency("hurt_enemy_cm"),
        snd_dependency("hurt_enemy_bullet"),
        snd_dependency("hurt_enemy_needler"),
        snd_dependency("hurt_enemy_plasma"),
        snd_dependency("hurt_enemy_sniper"),
        snd_dependency("hurt_enemy_grenade"),
        snd_dependency("hurt_enemy_explosion"),
        snd_dependency("hurt_enemy_melee"),
        snd_dependency("hurt_enemy_flame"),
        snd_dependency("hurt_enemy_shotgun"),
        snd_dependency("hurt_enemy_vehicle"),
        snd_dependency("hurt_enemy_mounted_weapon"),
        Pad(48),
        ),

    Struct("killing_people",
        snd_dependency("killed_friend"),
        snd_dependency("killed_friend_cm"),
        snd_dependency("killed_friend_player"),
        snd_dependency("killed_friend_player_cm"),
        snd_dependency("killed_enemy"),
        snd_dependency("killed_enemy_cm"),
        snd_dependency("killed_enemy_player"),
        snd_dependency("killed_enemy_player_cm"),
        snd_dependency("killed_enemy_covenant"),
        snd_dependency("killed_enemy_covenant_cm"),
        snd_dependency("killed_enemy_floodcombat"),
        snd_dependency("killed_enemy_floodcombat_cm"),
        snd_dependency("killed_enemy_floodcarrier"),
        snd_dependency("killed_enemy_floodcarrier_cm"),
        snd_dependency("killed_enemy_sentinel"),
        snd_dependency("killed_enemy_sentinel_cm"),

        snd_dependency("killed_enemy_bullet"),
        snd_dependency("killed_enemy_needler"),
        snd_dependency("killed_enemy_plasma"),
        snd_dependency("killed_enemy_sniper"),
        snd_dependency("killed_enemy_grenade"),
        snd_dependency("killed_enemy_explosion"),
        snd_dependency("killed_enemy_melee"),
        snd_dependency("killed_enemy_flame"),
        snd_dependency("killed_enemy_shotgun"),
        snd_dependency("killed_enemy_vehicle"),
        snd_dependency("killed_enemy_mounted_weapon"),
        snd_dependency("killing_spree"),
        Pad(48),
        ),

    Struct("player_kill_responses",
        snd_dependency("player_kill_cm"),
        snd_dependency("player_kill_bullet_cm"),
        snd_dependency("player_kill_needler_cm"),
        snd_dependency("player_kill_plasma_cm"),
        snd_dependency("player_kill_sniper_cm"),
        snd_dependency("anyone_kill_grenade_cm"),
        snd_dependency("player_kill_explosion_cm"),
        snd_dependency("player_kill_melee_cm"),
        snd_dependency("player_kill_flame_cm"),
        snd_dependency("player_kill_shotgun_cm"),
        snd_dependency("player_kill_vehicle_cm"),
        snd_dependency("player_kill_mounted_weapon_cm"),
        snd_dependency("player_killing_spree_cm"),
        Pad(48),
        ),

    Struct("friends_dying",
        snd_dependency("friend_died"),
        snd_dependency("friend_player_died"),
        snd_dependency("friend_killed_by_friend"),
        snd_dependency("friend_killed_by_friendly_player"),
        snd_dependency("friend_enemy"),
        snd_dependency("friend_enemy_player"),
        snd_dependency("friend_covenant"),
        snd_dependency("friend_flood"),
        snd_dependency("friend_sentinel"),
        snd_dependency("friend_betrayed"),
        Pad(32),
        ),

    Struct("shouting",
        snd_dependency("new_combat_alone"),
        snd_dependency("new_enemy_recent_combat"),
        snd_dependency("old_enemy_sighted"),
        snd_dependency("unexpected_enemy"),
        snd_dependency("dead_friend_found"),
        snd_dependency("alliance_broken"),
        snd_dependency("alliance_reformed"),
        snd_dependency("grenade_throwing"),
        snd_dependency("grenade_sighted"),
        snd_dependency("grenade_startle"),
        snd_dependency("grenade_danger_enemy"),
        snd_dependency("grenade_danger_self"),
        snd_dependency("grenade_danger_friend"),
        Pad(32),
        ),

    Struct("group_communication",
        snd_dependency("new_combat_group_re"),
        snd_dependency("new_combat_nearby_re"),
        snd_dependency("alert_friend"),
        snd_dependency("alert_friend_re"),
        snd_dependency("alert_lost_contact"),
        snd_dependency("alert_lost_contact_re"),
        snd_dependency("blocked"),
        snd_dependency("blocked_re"),
        snd_dependency("search_start"),
        snd_dependency("search_query"),
        snd_dependency("search_query_re"),
        snd_dependency("search_report"),
        snd_dependency("search_abandon"),
        snd_dependency("search_group_abandon"),
        snd_dependency("group_uncover"),
        snd_dependency("group_uncover_re"),
        snd_dependency("advance"),
        snd_dependency("advance_re"),
        snd_dependency("retreat"),
        snd_dependency("retreat_re"),
        snd_dependency("cover"),
        Pad(64),
        ),

    Struct("actions",
        snd_dependency("sighted_friend_player"),
        snd_dependency("shooting"),
        snd_dependency("shooting_vehicle"),
        snd_dependency("shooting_berserk"),
        snd_dependency("shooting_group"),
        snd_dependency("shooting_traitor"),
        snd_dependency("taunt"),
        snd_dependency("taunt_re"),
        snd_dependency("flee"),
        snd_dependency("flee_re"),
        snd_dependency("free_leader_died"),
        snd_dependency("attempted_flee"),
        snd_dependency("attempted_flee_re"),
        snd_dependency("lost_contact"),
        snd_dependency("hiding_finished"),
        snd_dependency("vehicle_entry"),
        snd_dependency("vehicle_exit"),
        snd_dependency("vehicle_woohoo"),
        snd_dependency("vehicle_scared"),
        snd_dependency("vehicle_collision"),
        snd_dependency("partially_sighted"),
        snd_dependency("nothing_there"),
        snd_dependency("pleading"),
        Pad(96),
        ),

    Struct("exclamations",
        snd_dependency("surprise"),
        snd_dependency("berserk"),
        snd_dependency("melee_attack"),
        snd_dependency("dive"),
        snd_dependency("uncover_exclamation"),
        snd_dependency("leap_attack"),
        snd_dependency("resurrection"),
        Pad(64),
        ),

    Struct("post_combat_actions",
        snd_dependency("celebration"),
        snd_dependency("check_body_enemy"),
        snd_dependency("check_body_friend"),
        snd_dependency("shooting_dead_enemy"),
        snd_dependency("shooting_dead_enemy_player"),
        Pad(64),
        ),

    Struct("post_combat_chatter",
        snd_dependency("alone"),
        snd_dependency("unscathed"),
        snd_dependency("seriously_wounded"),
        snd_dependency("seriously_wounded_re"),
        snd_dependency("massacre"),
        snd_dependency("massacre_re"),
        snd_dependency("rout"),
        snd_dependency("rout_re"),
        ),

    SIZE=4112,
    )

udlg_def = TagDef("udlg",
    blam_header('udlg'),
    udlg_body,

    ext=".dialogue", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.meta.wrappers.halo3_map import Halo3Map

class Halo5Map(Halo3Map):
    tag_defs_module = ""
    tag_classes_to_load = tuple()



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.h2.common_descs import *
from supyr_struct.defs.tag_def import TagDef


control_point = Struct("control point",
    QStruct("position", INCLUDE=ijk_float),
    QStruct("orientation", INCLUDE=ijkw_float),
    SIZE=28,
    )

trak_body = Struct("tagdata",
    Pad(4),
    h2_reflexive("control points", control_point, 16),
    ENDIAN="<", SIZE=12
    )


def get():
    return trak_def

trak_def = TagDef("trak",
    h2_blam_header('trak'),
    trak_body,

    ext=".%s" % h2_tag_class_fcc_to_ext["trak"], endian="<"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import sys

try:
    from reclaimer.sounds.ext import ogg_ext
    OGG_VORBIS_AVAILABLE = True
except ImportError:
    OGG_VORBIS_AVAILABLE = False

try:
    from reclaimer.sounds.ext import wma_ext
    WMA_AVAILABLE = True
except ImportError:
    WMA_AVAILABLE = False


SOUND_COMPILE_MODE_NEW = 0
SOUND_COMPILE_MODE_PRESERVE = 1
SOUND_COMPILE_MODE_ADDITIVE = 2


COMPRESSION_UNKNOWN = -1
# NOTE: the ordering of these constants is such that their endianess
# can be swapped by flipping the first bit. This is used in util.py
COMPRESSION_PCM_16_LE = 0  # 16bit signed pcm
COMPRESSION_PCM_16_BE = COMPRESSION_PCM_16_LE + 1
COMPRESSION_PCM_24_LE = 2  # 24bit signed pcm
COMPRESSION_PCM_24_BE = COMPRESSION_PCM_24_LE + 1
COMPRESSION_PCM_32_LE = 4  # 32bit signed pcm
COMPRESSION_PCM_32_BE = COMPRESSION_PCM_32_LE + 1

COMPRESSION_PCM_8_SIGNED   = 6
COMPRESSION_PCM_8_UNSIGNED = 7

COMPRESSION_XBOX_ADPCM = 16
COMPRESSION_IMA_ADPCM  = 17
COMPRESSION_OGG   = 18  # halo pc only
COMPRESSION_WMA   = 19  # halo 2 only

# these encoding constants mirror halo 1/2 enum values.
ENCODING_UNKNOWN = -1
ENCODING_MONO   = 0
ENCODING_STEREO = 1
ENCODING_CODEC  = 2

SAMPLE_RATE_22K = 22050
SAMPLE_RATE_32K = 32000
SAMPLE_RATE_44K = 44100

# Halo constants
DEFAULT_PITCH_RANGE_NAME = "default"

SAMPLE_RATE_MOUTH_DATA = 30

MAX_SAMPLE_CHUNK_SIZE = 0x400000
MAX_MOUTH_DATA        = 0x2000
DEF_SAMPLE_CHUNK_SIZE = 0x38E00  # the number of bytes of pcm data
#                                  to split input samples into when
#                                  splitting into smaller chunks

MAX_HALO_SAMPLE_RATE = 48000

XBOX_ADPCM_COMPRESSED_BLOCKSIZE   = 36
XBOX_ADPCM_DECOMPRESSED_BLOCKSIZE = 128
IMA_ADPCM_COMPRESSED_BLOCKSIZE   = 36   # not correct
IMA_ADPCM_DECOMPRESSED_BLOCKSIZE = 128  # not correct

# Wave file format constants
WAV_FORMAT_PCM        = 0x0001
WAV_FORMAT_PCM_FLOAT  = 0x0003
WAV_FORMAT_IMA_ADPCM  = 0x0011
WAV_FORMAT_XBOX_ADPCM = 0x0069

ALLOWED_WAV_FORMATS = set((
    WAV_FORMAT_PCM,
    WAV_FORMAT_PCM_FLOAT,
    WAV_FORMAT_IMA_ADPCM,
    WAV_FORMAT_XBOX_ADPCM
    ))

# Endianness interop constants
if sys.byteorder == "little":
    ADPCM_DECOMPRESSED_FORMAT = COMPRESSION_PCM_16_LE
    DEFAULT_UNCOMPRESSED_FORMAT = COMPRESSION_PCM_16_LE
    HIGHEST_FIDELITY_FORMAT = COMPRESSION_PCM_32_LE
    NATIVE_ENDIANNESS_FORMATS = {
        COMPRESSION_PCM_16_LE, COMPRESSION_PCM_24_LE,
        COMPRESSION_PCM_32_LE,
        COMPRESSION_PCM_8_SIGNED, COMPRESSION_PCM_8_UNSIGNED
        }
else:
    ADPCM_DECOMPRESSED_FORMAT = COMPRESSION_PCM_16_BE
    DEFAULT_UNCOMPRESSED_FORMAT = COMPRESSION_PCM_16_BE
    HIGHEST_FIDELITY_FORMAT = COMPRESSION_PCM_32_BE
    NATIVE_ENDIANNESS_FORMATS = {
        COMPRESSION_PCM_16_BE, COMPRESSION_PCM_24_BE,
        COMPRESSION_PCM_32_BE,
        COMPRESSION_PCM_8_SIGNED, COMPRESSION_PCM_8_UNSIGNED
        }

PCM_FORMATS = {
    COMPRESSION_PCM_16_LE, COMPRESSION_PCM_16_BE,
    COMPRESSION_PCM_24_LE, COMPRESSION_PCM_24_BE,
    COMPRESSION_PCM_32_LE, COMPRESSION_PCM_32_BE,
    COMPRESSION_PCM_8_SIGNED, COMPRESSION_PCM_8_UNSIGNED
    }

channel_counts = {
    ENCODING_UNKNOWN: 1,
    ENCODING_MONO:    1,
    ENCODING_STEREO:  2,
    ENCODING_CODEC:   6,
    }

sample_widths = {
    COMPRESSION_PCM_8_SIGNED: 1,
    COMPRESSION_PCM_8_UNSIGNED: 1,
    COMPRESSION_PCM_16_LE: 2,
    COMPRESSION_PCM_16_BE: 2,
    COMPRESSION_PCM_24_LE: 3,
    COMPRESSION_PCM_24_BE: 3,
    COMPRESSION_PCM_32_LE: 4,
    COMPRESSION_PCM_32_BE: 4,
    }

# maps wave format enum options and sample widths to our compression constants
wav_format_mapping = {
    (WAV_FORMAT_PCM, 1): COMPRESSION_PCM_8_UNSIGNED,
    (WAV_FORMAT_PCM, 2): COMPRESSION_PCM_16_LE,
    (WAV_FORMAT_PCM, 3): COMPRESSION_PCM_24_LE,
    (WAV_FORMAT_PCM, 4): COMPRESSION_PCM_32_LE,
    (WAV_FORMAT_IMA_ADPCM, None): COMPRESSION_IMA_ADPCM,
    (WAV_FORMAT_XBOX_ADPCM, None): COMPRESSION_XBOX_ADPCM,
    }

# these mappings key halo 1 compression enums
# to the compression/sample rate constants
halo_1_compressions = {
    0: COMPRESSION_PCM_16_LE,
    1: COMPRESSION_XBOX_ADPCM,
    2: COMPRESSION_IMA_ADPCM,
    3: COMPRESSION_OGG,
    }
halo_1_sample_rates = {
    0: SAMPLE_RATE_22K,
    1: SAMPLE_RATE_44K,
    }

# these mappings key halo 2 compression enums
# to the compression/sample rate constants
halo_2_compressions = {
    0: COMPRESSION_PCM_16_BE,
    1: COMPRESSION_XBOX_ADPCM,
    2: COMPRESSION_IMA_ADPCM,
    3: COMPRESSION_PCM_16_LE,
    4: COMPRESSION_WMA,
    }
halo_2_sample_rates = {
    0: SAMPLE_RATE_22K,
    1: SAMPLE_RATE_44K,
    2: SAMPLE_RATE_32K,
    }

# unneeded for export
del sys



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import lzma
import zlib
import os

from copy import deepcopy

from reclaimer.constants import map_build_dates, map_magics, GEN_3_ENGINES
from reclaimer.meta.halo1_map import map_header_def, map_header_vap_def,\
     map_header_anni_def, map_header_demo_def, tag_index_pc_def,\
     tag_index_xbox_def, tag_index_anni_def
from reclaimer.meta.halo2_alpha_map import h2_alpha_map_header_def,\
     h2_alpha_tag_index_def
from reclaimer.meta.halo2_map import h2v_map_header_full_def,\
     h2v_map_header_def, h2x_map_header_full_def, h2x_map_header_def,\
     h2_tag_index_def
from reclaimer.meta.halo3_map import h3_map_header_def, h3_tag_index_def
from reclaimer.meta.shadowrun_map import sr_tag_index_def
from reclaimer.meta.stubbs_map import stubbs_tag_index_def

from supyr_struct.defs.tag_def import TagDef
from supyr_struct.buffer import get_rawdata


def get_map_version(header):
    try:
        version = header.version.enum_name
    except AttributeError:
        return None
    try:
        build_date = header.build_date
    except AttributeError:
        build_date = None

    if version == "halo1pc":
        # apparently the halo pc build and halo 2 alpha
        # use the same version integer of 7
        if build_date == map_build_dates["halo2alpha"]:
            version = "halo2alpha"
        elif build_date == map_build_dates["halo1anni"]:
            version = "halo1anni"
    elif version == "halo1xbox":
        if build_date is None:
            version = None
        elif build_date == map_build_dates["stubbs"]:
            version = "stubbs"
        elif build_date == "":
            if header.unknown in (11, 1033):
                # this is the only discernable difference
                # between xbox demo and stubbs pc maps
                version = "halo1xboxdemo"
            else:
                version = "stubbspc"
        elif build_date == map_build_dates["shadowrun_proto"]:
            version = "shadowrun_proto"
    elif hasattr(header, "yelo_header") and (
            header.yelo_header.yelo.enum_name == "yelo"):
        version = "halo1yelo"
    elif version == "halo2":
        version = None
        if build_date == map_build_dates['halo2beta']:
            version = "halo2beta"
        elif build_date == map_build_dates['halo2xbox']:
            version = "halo2xbox"
        elif build_date == map_build_dates['halo2epsilon']:
            version = "halo2epsilon"
        elif build_date == map_build_dates['halo2vista']:
            version = "halo2vista"
    elif version == "halo3":
        if build_date == map_build_dates['halo3odst']:
            version = "halo3odst"
    elif version == "haloreach":
        if build_date == map_build_dates['haloreachbeta']:
            version = "haloreachbeta"
        elif build_date == map_build_dates['halo4']:
            version = "halo4"

    return version


def get_map_header(map_file, header_only=False):
    if hasattr(map_file, "read"):
        orig_pos = map_file.tell()
        map_file.seek(0)
        header_data = map_file.read(2048)
        map_file.seek(orig_pos)
    else:
        header_data = map_file

    sig   = header_data[:4]
    sig_big    = sig.decode('latin-1').lower()
    sig_little = sig_big[::-1]
    ver_big    = int.from_bytes(header_data[4:8], 'big')
    ver_little = int.from_bytes(header_data[4:8], 'little')
    header_def = None

    if sig_little == "head":
        if ver_little == 8:
            build_date = header_data[300: 324].decode("latin-1")
            if build_date == map_build_dates['halo2vista']:
                header_def = h2v_map_header_full_def
                if header_only:
                    header_def = h2v_map_header_def
            else:
                header_def = h2x_map_header_full_def
                if header_only:
                    header_def = h2x_map_header_def

        elif ver_little in (5, 6, 609):
            header_def = map_header_def

        elif ver_little == 134:
            header_def = map_header_vap_def

        elif ver_little == 7:
            header_def = map_header_def
            build_date = header_data[64: 96].decode("latin-1")
            if build_date == map_build_dates['halo2alpha']:
                header_def = h2_alpha_map_header_def

    elif sig_big == "head":
        if ver_big == 11:
            header_def = h3_map_header_def
        elif ver_big == 7:
            header_def = map_header_anni_def

    elif header_data[704:708] == b'dehE' and header_data[1520:1524] == b'tofG':
        header_def = map_header_demo_def

    elif header_data[:2] == b'\x78\xDA':
        # zlib compressed halo 2 vista map.
        # decompress the 2048 byte header
        header_def = h2v_map_header_full_def
        decomp_len = None
        if header_only:
            header_def = h2v_map_header_def
            decomp_len = 2048

        map_file = zlib.decompressobj().decompress(map_file, decomp_len)

    if header_def is None:
        return None

    return header_def.build(rawdata=map_file)


def get_tag_index(map_data, header=None):
    if header is None:
        header = get_map_header(map_data)

    if header is None:
        return None

    map_data = decompress_map(map_data, header)
    base_address = header.tag_index_header_offset

    tag_index_def = tag_index_pc_def
    version = get_map_version(header)
    if "shadowrun" in version:
        tag_index_def = sr_tag_index_def
    elif "stubbs" in version:
        tag_index_def = stubbs_tag_index_def
    elif header.version.data < 6:
        tag_index_def = tag_index_xbox_def
    elif version == "halo2alpha":
        tag_index_def = h2_alpha_tag_index_def
    elif version == "halo1anni":
        tag_index_def = tag_index_anni_def
    elif header.version.enum_name == "halo2":
        tag_index_def = h2_tag_index_def
    elif header.version.enum_name == "halo3":
        tag_index_def = h3_tag_index_def

        for partition in header.partitions:
            if base_address in range(partition.load_address,
                                     partition.load_address + partition.size):
                base_address -= partition.load_address - partition.file_offset
                break

    if hasattr(header, "sections") and header.sections.tag.size == 0:
        # shared resource map. no tag index, so make a blank one
        return tag_index_def.build(map_header=header)

    tag_index = tag_index_def.build(
        rawdata=map_data, magic=get_map_magic(header),
        offset=base_address, map_header=header)

    return tag_index


def get_index_magic(header):
    version = get_map_version(header)
    if version == "halo2vista":
        return header.virtual_address
    elif version in GEN_3_ENGINES:
        base_address = header.tag_index_header_offset
        for partition in header.partitions:
            if base_address in range(partition.load_address,
                                     partition.load_address + partition.size):
                return partition.load_address - partition.file_offset

    return map_magics.get(version, 0)


def get_map_magic(header):
    magic = get_index_magic(header)
    if header.version.enum_name not in GEN_3_ENGINES:
        magic -= header.tag_index_header_offset

    return magic


def get_is_compressed_map(comp_data, header):
    if header.version.data == 134:
        return header.vap_header.compression_type.data != 0
    elif header.version.data not in (7, 609):
        decomp_len = header.decomp_len
        if get_map_version(header) == "pcstubbs":
            decomp_len -= 2048

        return decomp_len > len(comp_data)
    return False


def decompress_map(comp_data, header=None, decomp_path=None, writable=False):
    if header is None:
        header = get_map_header(comp_data)

    if not get_is_compressed_map(comp_data, header):
        # not actually compressed
        return comp_data
    elif header.version.data == 134:
        if header.vap_header.compression_type.enum_name == "lzma":
            return decompress_map_lzma(comp_data, header, decomp_path, writable)
        else:
            return comp_data
    else:
        return decompress_map_deflate(comp_data, header, decomp_path, writable)


def decompress_map_lzma(comp_data, header, decomp_path="", writable=False):
    vap_header = header.vap_header

    os.makedirs(os.path.dirname(decomp_path), exist_ok=True)
    if not decomp_path:
        decomp_path = "decomp.map"

    # get the info needed to decompress each block
    comp_offs = list(b.file_offset for b in vap_header.blocks)
    comp_sizes = list(b.file_size for b in vap_header.blocks)
    if not comp_offs:
        # if blocks is 0, entire compressed size is assumed to be one stream
        comp_offs.append(header.SIZE)
        comp_sizes.append(vap_header.compressed_size - header.SIZE)

    # copy the header and change the compression fields
    header_copy = deepcopy(header)
    header_copy.vap_header.compression_type.set_to("uncompressed")
    header_copy.vap_header.block_count = 0
    header_copy.vap_header.compressed_size = 0
    del header_copy.vap_header.blocks[:]

    with open(decomp_path, "wb+") as f:
        # size the uncompressed file to the right size
        f.seek(vap_header.decompressed_size - 1)
        f.write(b"\x00")

        # write the copied header to the map
        f.seek(0)
        f.write(header_copy.serialize())

        for in_off, in_size in zip(comp_offs, comp_sizes):
            comp_data.seek(in_off)
            f.write(lzma.decompress(comp_data.read(in_size)))

    # have to do this separate or seeking will be fucked
    return get_rawdata(filepath=decomp_path, writable=writable)


def decompress_map_deflate(comp_data, header, decomp_path="", writable=False):
    comp_data.seek(0)
    decomp_start = 2048
    decomp_len   = header.decomp_len
    version      = get_map_version(header)
    if version == "pcstubbs":
        decomp_len -= 2048
    elif version == "halo2vista":
        decomp_start = 0

    os.makedirs(os.path.dirname(decomp_path), exist_ok=True)
    if not decomp_path:
        decomp_path = "decomp.map"

    with open(decomp_path, "wb+") as f:
        if decomp_start:
            comp_data.seek(0)
            f.write(comp_data.read(decomp_start))
            comp_data.seek(decomp_start)
            comp_data = comp_data.read()

        decomp_obj = zlib.decompressobj()
        last_comp_len = len(comp_data)
        while comp_data:
            # decompress map 64Mb at a time
            f.write(decomp_obj.decompress(comp_data, 64*1024*1024))
            comp_data = decomp_obj.unconsumed_tail
            if len(comp_data) == last_comp_len:
                # need to do this to not get stuck in an infinite
                # loop where the compressed data doesnt actually
                # decode to anything because it is all zeros.
                break
            last_comp_len = len(comp_data)

        # pad the file to its decompressed length
        f.write(b'\xca'*(decomp_len - f.tell()))

    # have to do this separate or seeking will be fucked
    return get_rawdata(filepath=decomp_path, writable=writable)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


crte_unknown_6_unknown_0 = Struct("unknown_0",
    BytesRaw("unknown", SIZE=32, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


crte_unknown_6_unknown_1_unknown_1_unknown_0 = Struct("unknown_0",
    BytesRaw("unknown", SIZE=16, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


crte_unknown_6_unknown_1_unknown_1 = Struct("unknown_1",
    h3_reflexive("unknown_0", crte_unknown_6_unknown_1_unknown_1_unknown_0),
    h3_rawdata_ref("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


crte_unknown_6_unknown_1 = Struct("unknown_1",
    BytesRaw("unknown_0", SIZE=28, VISIBLE=False),
    h3_reflexive("unknown_1", crte_unknown_6_unknown_1_unknown_1),
    VISIBLE=False,
    ENDIAN=">", SIZE=40
    )


crte_unknown_6 = Struct("unknown_6",
    h3_reflexive("unknown_0", crte_unknown_6_unknown_0),
    h3_reflexive("unknown_1", crte_unknown_6_unknown_1),
    VISIBLE=False,
    ENDIAN=">", SIZE=24
    )


crte_unknown_7_unknown_0_unknown_1_unknown_0 = Struct("unknown_0",
    BytesRaw("unknown", SIZE=16, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


crte_unknown_7_unknown_0_unknown_1 = Struct("unknown_1",
    h3_reflexive("unknown_0", crte_unknown_7_unknown_0_unknown_1_unknown_0),
    h3_rawdata_ref("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


crte_unknown_7_unknown_0 = Struct("unknown_0",
    BytesRaw("unknown_0", SIZE=28, VISIBLE=False),
    h3_reflexive("unknown_1", crte_unknown_7_unknown_0_unknown_1),
    BytesRaw("unknown_2", SIZE=12, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=52
    )


crte_unknown_7_unknown_1 = Struct("unknown_1",
    BytesRaw("unknown", SIZE=52, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=52
    )


crte_unknown_7_unknown_2_unknown_1_unknown_0 = Struct("unknown_0",
    BytesRaw("unknown", SIZE=12, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=12
    )


crte_unknown_7_unknown_2_unknown_1 = Struct("unknown_1",
    h3_reflexive("unknown_0", crte_unknown_7_unknown_2_unknown_1_unknown_0),
    h3_rawdata_ref("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


crte_unknown_7_unknown_2 = Struct("unknown_2",
    BytesRaw("unknown_0", SIZE=24, VISIBLE=False),
    h3_reflexive("unknown_1", crte_unknown_7_unknown_2_unknown_1),
    BytesRaw("unknown_2", SIZE=12, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=48
    )


crte_unknown_7_unknown_3_unknown_1_unknown_0 = Struct("unknown_0",
    BytesRaw("unknown", SIZE=16, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


crte_unknown_7_unknown_3_unknown_1 = Struct("unknown_1",
    h3_reflexive("unknown_0", crte_unknown_7_unknown_3_unknown_1_unknown_0),
    h3_rawdata_ref("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


crte_unknown_7_unknown_3 = Struct("unknown_3",
    BytesRaw("unknown_0", SIZE=28, VISIBLE=False),
    h3_reflexive("unknown_1", crte_unknown_7_unknown_3_unknown_1),
    BytesRaw("unknown_2", SIZE=12, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=52
    )


crte_unknown_7 = Struct("unknown_7",
    h3_reflexive("unknown_0", crte_unknown_7_unknown_0),
    h3_reflexive("unknown_1", crte_unknown_7_unknown_1),
    h3_reflexive("unknown_2", crte_unknown_7_unknown_2),
    h3_reflexive("unknown_3", crte_unknown_7_unknown_3),
    VISIBLE=False,
    ENDIAN=">", SIZE=48
    )


crte_unknown_9_unknown_0 = Struct("unknown_0",
    BytesRaw("unknown", SIZE=48, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=48
    )


crte_unknown_9_unknown_1_unknown_1_unknown_0 = Struct("unknown_0",
    BytesRaw("unknown", SIZE=16, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


crte_unknown_9_unknown_1_unknown_1 = Struct("unknown_1",
    h3_reflexive("unknown_0", crte_unknown_9_unknown_1_unknown_1_unknown_0),
    h3_rawdata_ref("unknown_1", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


crte_unknown_9_unknown_1 = Struct("unknown_1",
    BytesRaw("unknown_0", SIZE=28, VISIBLE=False),
    h3_reflexive("unknown_1", crte_unknown_9_unknown_1_unknown_1),
    BytesRaw("unknown_2", SIZE=12, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=52
    )


crte_unknown_9_unknown_2 = Struct("unknown_2",
    BytesRaw("unknown", SIZE=12, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=12
    )


crte_unknown_9 = Struct("unknown_9",
    h3_reflexive("unknown_0", crte_unknown_9_unknown_0),
    h3_reflexive("unknown_1", crte_unknown_9_unknown_1),
    h3_reflexive("unknown_2", crte_unknown_9_unknown_2),
    VISIBLE=False,
    ENDIAN=">", SIZE=36
    )


crte_body = Struct("tagdata",
    h3_string_id("unknown_0", VISIBLE=False),
    BytesRaw("unknown_1", SIZE=16, VISIBLE=False),
    h3_rawdata_ref("unknown_2", VISIBLE=False),
    h3_rawdata_ref("unknown_3", VISIBLE=False),
    h3_dependency("unknown_4", VISIBLE=False),
    h3_string_id("unknown_5", VISIBLE=False),
    h3_reflexive("unknown_6", crte_unknown_6),
    h3_reflexive("unknown_7", crte_unknown_7),
    BytesRaw("unknown_8", SIZE=12, VISIBLE=False),
    h3_reflexive("unknown_9", crte_unknown_9),
    ENDIAN=">", SIZE=128
    )


def get():
    return crte_def

crte_def = TagDef("crte",
    h3_blam_header('crte'),
    crte_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["crte"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os

from reclaimer.model.jms import JmsVertex
from reclaimer.hek.defs.scex import scex_def
from reclaimer.hek.defs.schi import schi_def
from reclaimer.hek.defs.senv import senv_def
from reclaimer.hek.defs.sgla import sgla_def
from reclaimer.hek.defs.smet import smet_def
from reclaimer.hek.defs.soso import soso_def
from reclaimer.hek.defs.sotr import sotr_def
from reclaimer.hek.defs.spla import spla_def
from reclaimer.hek.defs.swat import swat_def
from reclaimer.hek.defs.mod2 import triangle as mod2_tri_struct, \
     fast_uncompressed_vertex as mod2_vert_struct
from reclaimer.common_descs import raw_reflexive, BlockDef

__all__ = (
    'mod2_verts_def', 'mod2_tri_strip_def',
    'LOD_NAMES', 'MAX_STRIP_LEN', 'EMPTY_GEOM_VERTS',
    'generate_shader',
    )


mod2_verts_def = BlockDef(
    raw_reflexive("vertices", mod2_vert_struct, 65535),
    endian='>'
    )

mod2_tri_strip_def = BlockDef(
    raw_reflexive("triangle", mod2_tri_struct, 65535),
    endian='>'
    )

LOD_NAMES = ("superhigh", "high", "medium", "low", "superlow")
MAX_STRIP_LEN = 32763 * 3

EMPTY_GEOM_VERTS = (
    JmsVertex(0, 0.000000001, 0.0, 0.0,
              0.0, 0.0, 1.0,
              -1, 0.0, 0.0, 0.0),
    JmsVertex(0, 0.0, 0.000000001, 0.0,
              0.0, 0.0, 1.0,
              -1, 0.0, 0.0, 1.0),
    JmsVertex(0, 0.0, 0.0, 0.000000001,
              0.0, 0.0, 1.0,
              -1, 0.0, 1.0, 0.0),
    )

def generate_shader(jms_material, tags_dir, data_dir=""):
    shdr_type = jms_material.shader_type
    shdr_path = jms_material.shader_path

    if not shdr_path:
        return

    tag_path = "%s.%s" % (os.path.join(tags_dir, shdr_path), shdr_type)
    if os.path.isfile(tag_path):
        # dont make shaders that already exist
        return

    shdr_blockdef = None
    if shdr_type == "shader_transparent_chicago_extended":
        shdr_blockdef = scex_def
    elif shdr_type == "shader_transparent_chicago":
        shdr_blockdef = schi_def
    elif shdr_type == "shader_environment":
        shdr_blockdef = senv_def
    elif shdr_type == "shader_glass":
        shdr_blockdef = sgla_def
    elif shdr_type == "shader_meter":
        shdr_blockdef = smet_def
    elif shdr_type == "shader_model":
        shdr_blockdef = soso_def
    elif shdr_type == "shader_transparent_generic":
        shdr_blockdef = sotr_def
    elif shdr_type == "shader_plasma":
        shdr_blockdef = spla_def
    elif shdr_type == "shader_water":
        shdr_blockdef = swat_def
    else:
        return

    bitmap_path = ""
    if jms_material.tiff_path not in ('', '<none>') and data_dir:
        try:
            # TODO: Make this posix compat
            bitmap_path = os.path.relpath(
                jms_material.tiff_path.replace("/", "\\"),
                data_dir.replace("/", "\\")).strip(" ")
        except Exception:
            pass

        if bitmap_path.startswith("."):
            bitmap_path = ""

    shdr_tag = shdr_blockdef.build()
    shdr_tag.filepath = tag_path
    tag_data = shdr_tag.data.tagdata

    if not bitmap_path or bitmap_path.lower() == "<none>":
        pass
    elif "chicago" in shdr_type:
        if shdr_type == "shader_transparent_chicago":
            maps = tag_data.schi_attrs.maps
        else:
            maps = tag_data.scex_attrs.four_stage_maps
        maps.STEPTREE.append()
        map = maps.STEPTREE[-1]
        map.bitmap.map_u_scale = map.bitmap.map_v_scale = 1.0
        map.bitmap.filepath = bitmap_path
    elif "environment" in shdr_type:
        tag_data.senv_attrs.diffuse.base_map.filepath = bitmap_path
    elif "glass" in shdr_type:
        tag_data.sgla_attrs.diffuse_properties.map.filepath = bitmap_path
    elif "meter" in shdr_type:
        tag_data.smet_attrs.meter_shader.map.filepath = bitmap_path
    elif "model" in shdr_type:
        tag_data.soso_attrs.maps.diffuse_map.filepath = bitmap_path
    elif "water" in shdr_type:
        tag_data.swat_attrs.water_shader.base_map.filepath = bitmap_path

    shdr_tag.serialize(temp=False, calc_pointers=False, int_test=False)

    return



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

'''
Most byteswapping is handeled by supyr_struct by changing the endianness,
but certain chunks of raw data are significantly faster to just write
byteswapping routines for, like raw vertex, triangle, and animation data.
'''
import array

from reclaimer.sounds.util import byteswap_pcm16_sample_data
from supyr_struct.field_types import BytearrayRaw
from supyr_struct.defs.block_def import BlockDef

try:
    from .ext import byteswapping_ext
    fast_byteswapping = True
except:
    fast_byteswapping = False


raw_block_def = BlockDef("raw_block",
    BytearrayRaw('data',
        SIZE=lambda node, *a, **kw: 0 if node is None else len(node))
    )

def make_mutable_struct_array_copy(data, struct_size):
    valid_length = struct_size*(len(data)//struct_size)
    if valid_length == len(data):
        return bytearray(data)
    return bytearray(data[:valid_length])


def byteswap_struct_array(original, swapped, size, count=None, start=0,
                          two_byte_offs=(), four_byte_offs=(), eight_byte_offs=()):
    assert start >= 0

    for off in two_byte_offs:
        assert (off + 2) <= size

    for off in four_byte_offs:
        assert (off + 4) <= size

    for off in eight_byte_offs:
        assert (off + 8) <= size

    if size <= 0:
        return

    max_count = (min(len(original), len(swapped)) - start) // size
    if count is None:
        count = max_count
    else:
        count = min(max_count, count)

    if count <= 0:
        return

    end = start + count * size
    if fast_byteswapping:
        byteswapping_ext.byteswap_struct_array(
            original, swapped, start, end, size,
            array.array("I", two_byte_offs),
            array.array("I", four_byte_offs),
            array.array("I", eight_byte_offs),
            )
        return

    for field_off in two_byte_offs:
        for off in range(field_off + start, end, size):
            swapped[off]     = original[off + 1]
            swapped[off + 1] = original[off]

    for field_off in four_byte_offs:
        for off in range(field_off + start, end, size):
            swapped[off]     = original[off + 3]
            swapped[off + 1] = original[off + 2]
            swapped[off + 2] = original[off + 1]
            swapped[off + 3] = original[off]

    for field_off in eight_byte_offs:
        for off in range(field_off + start, end, size):
            swapped[off]     = original[off + 7]
            swapped[off + 1] = original[off + 6]
            swapped[off + 2] = original[off + 5]
            swapped[off + 3] = original[off + 4]
            swapped[off + 4] = original[off + 3]
            swapped[off + 5] = original[off + 2]
            swapped[off + 6] = original[off + 1]
            swapped[off + 7] = original[off]


def byteswap_raw_reflexive(refl):
    desc = refl.desc
    struct_size, two_byte_offs, four_byte_offs = desc.get(
        "RAW_REFLEXIVE_INFO", (0, (), ()))
    if not two_byte_offs and not four_byte_offs:
        return

    original = refl.STEPTREE
    swapped = make_mutable_struct_array_copy(original, struct_size)
    byteswap_struct_array(original, swapped, struct_size, refl.size, 0,
                          two_byte_offs, four_byte_offs)
    refl.STEPTREE = swapped


def byteswap_coll_bsp(bsp):
    for b in bsp:
        byteswap_raw_reflexive(b)


def byteswap_pcm16_samples(pcm_block):
    # replace the verts with the byteswapped ones
    pcm_block.STEPTREE = bytearray(
        byteswap_pcm16_sample_data(pcm_block.STEPTREE))


def byteswap_sbsp_meta(meta):
    if len(meta.collision_bsp.STEPTREE):
        for b in meta.collision_bsp.STEPTREE[0]:
            byteswap_raw_reflexive(b)

    # do NOT need to swap meta.nodes since they are always little endian
    for b in (meta.leaves, meta.leaf_surfaces, meta.surface,
              meta.lens_flare_markers, meta.breakable_surfaces, meta.markers):
        byteswap_raw_reflexive(b)


def byteswap_scnr_script_syntax_data(meta):
    original = meta.script_syntax_data.data
    swapped = original[: ((len(original)-56)//20) * 20 + 56]

    # swap the 56 byte header
    # first 32 bytes are a string
    byteswap_struct_array(
        original, swapped, 56, 1, 0,
        two_byte_offs=(32, 34, 38, 44, 46, 48, 50),
        four_byte_offs=(40, 52))

    # swap the 20 byte blocks
    byteswap_struct_array(
        original, swapped, 20, None, 56,
        two_byte_offs=(0, 2, 4, 6),
        four_byte_offs=(8, 12, 16)
        )

    meta.script_syntax_data.data = swapped


def byteswap_uncomp_verts(verts_block):
    original = verts_block.STEPTREE.data
    swapped = make_mutable_struct_array_copy(original, 68)

    byteswap_struct_array(
        original, swapped, 68, None, 0,
        two_byte_offs=(56, 58),
        four_byte_offs=(0, 4, 8, 12, 16, 20, 24, 28, 32,
                        36, 40, 44, 48, 52, 60, 64)
        )

    verts_block.STEPTREE.data = swapped
    verts_block.size = len(swapped)//68


def byteswap_comp_verts(verts_block):
    original = verts_block.STEPTREE.data
    swapped = make_mutable_struct_array_copy(original, 32)

    byteswap_struct_array(
        original, swapped, 32, None, 0,
        two_byte_offs=(24, 26, 28, 30),
        four_byte_offs=(0, 4, 8, 12, 16, 20)
        )

    verts_block.STEPTREE.data = swapped
    verts_block.size = len(swapped)//32


def byteswap_tris(tris_block):
    original = tris_block.STEPTREE.data
    swapped = make_mutable_struct_array_copy(original, 6)

    byteswap_struct_array(
        original, swapped, 6, None, 0,
        two_byte_offs=(0, 2, 4)
        )

    tris_block.STEPTREE.data = swapped
    tris_block.size = len(swapped)//6


def byteswap_animation(anim):
    frame_info   = anim.frame_info.STEPTREE
    default_data = anim.default_data.STEPTREE
    frame_data   = anim.frame_data.STEPTREE

    comp_data_offset = anim.offset_to_compressed_data
    frame_count = anim.frame_count
    node_count  = anim.node_count
    trans_int = anim.trans_flags0 + (anim.trans_flags1<<32)
    rot_int   = anim.rot_flags0   + (anim.rot_flags1  <<32)
    scale_int = anim.scale_flags0 + (anim.scale_flags1<<32)

    trans_flags = tuple(bool(trans_int & (1 << i)) for i in range(node_count))
    rot_flags   = tuple(bool(rot_int   & (1 << i)) for i in range(node_count))
    scale_flags = tuple(bool(scale_int & (1 << i)) for i in range(node_count))

    frame_info_size = {0: 0, 1: 8, 2: 12, 3: 16}.get(
        anim.frame_info_type.data, 0) * frame_count
    frame_size = (12 * sum(trans_flags) +
                  8  * sum(rot_flags) +
                  4  * sum(scale_flags))
    default_data_size = node_count * (12 + 8 + 4) - frame_size
    uncomp_frame_data_size = frame_size * frame_count

    if len(frame_info) < frame_info_size:
        raise ValueError("Expected %s bytes of frame info in '%s', but got %s" %
                         (frame_info_size, anim.name, len(frame_info)))
    elif default_data and len(default_data) < default_data_size:
        raise ValueError("Expected %s bytes of default data in '%s', but got %s" %
                         (default_data_size, anim.name, len(default_data)))
    elif not anim.flags.compressed_data:
        if len(frame_data) - comp_data_offset < uncomp_frame_data_size:
            raise ValueError(
                "Expected %s bytes of frame data in '%s', but got %s" %
                (uncomp_frame_data_size, anim.name, len(frame_data)))

    new_frame_info   = bytearray(frame_info_size)
    new_default_data = bytearray(default_data_size)

    # some tags actually have the offset as non-zero in meta form
    # and it actually matters, so we need to take this into account
    new_uncomp_frame_data = bytearray(uncomp_frame_data_size)

    # byteswap the frame info
    byteswap_struct_array(
        frame_info, new_frame_info, 4, None, 0,
        four_byte_offs=(0, )
        )

    default_data_two_byte_offsets = ()
    default_data_four_byte_offsets = ()
    frame_data_two_byte_offsets = ()
    frame_data_four_byte_offsets = ()
    i = j = 0
    for n in range(node_count):
        if rot_flags[n]:
            frame_data_two_byte_offsets += tuple(range(i, i + 8, 2))
            i += 8
        else:
            default_data_two_byte_offsets += tuple(range(j, j + 8, 2))
            j += 8

        if trans_flags[n]:
            frame_data_four_byte_offsets += tuple(range(i, i + 12, 4))
            i += 12
        else:
            default_data_four_byte_offsets += tuple(range(j, j + 12, 4))
            j += 12

        if scale_flags[n]:
            frame_data_four_byte_offsets += (i, )
            i += 4
        else:
            default_data_four_byte_offsets += (j, )
            j += 4

    if default_data and default_data_size:
        # byteswap the default_data
        byteswap_struct_array(
            default_data, new_default_data, default_data_size, 1, 0,
            two_byte_offs=default_data_two_byte_offsets,
            four_byte_offs=default_data_four_byte_offsets,
            )

    if not anim.flags.compressed_data or comp_data_offset and frame_size:
        # byteswap the frame_data
        byteswap_struct_array(
            frame_data, new_uncomp_frame_data, frame_size, frame_count, 0,
            two_byte_offs=frame_data_two_byte_offsets,
            four_byte_offs=frame_data_four_byte_offsets,
            )

    anim.frame_info.STEPTREE   = new_frame_info
    anim.default_data.STEPTREE = new_default_data
    anim.frame_data.STEPTREE   = new_uncomp_frame_data
    anim.offset_to_compressed_data = 0

    if anim.flags.compressed_data:
        anim.offset_to_compressed_data = len(new_uncomp_frame_data)
        anim.frame_data.STEPTREE += frame_data[comp_data_offset:]



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.obje import ObjeTag
from supyr_struct.defs.tag_def import TagDef

def get():
    return obje_def

attachment = Struct('attachment',
    dependency('type', valid_attachments),
    ascii_str32('marker'),
    SEnum16('primary_scale', *function_outputs),
    SEnum16('secondary_scale', *function_outputs),
    SEnum16('change_color', *function_names),

    SIZE=72
    )

widget = Struct('widget',
    dependency('reference', valid_widgets),
    SIZE=32
    )

function = Struct('function',
    Bool32('flags',
        'invert',
        'additive',
        'always_active',
        ),
    float_sec('period', UNIT_SCALE=sec_unit_scale),  # seconds
    SEnum16('scale_period_by', *function_inputs_outputs),
    SEnum16('function', *animation_functions),
    SEnum16('scale_function_by', *function_inputs_outputs),
    SEnum16('wobble_function', *animation_functions),
    float_sec('wobble_period', UNIT_SCALE=sec_unit_scale),  # seconds
    Float('wobble_magnitude', SIDETIP="%"),  # percent

    Float('square_wave_threshold'),
    SInt16('step_count'),
    SEnum16('map_to', *fade_functions),
    SInt16('sawtooth_count'),
    SEnum16('add', *function_inputs_outputs),
    SEnum16('scale_result_by', *function_inputs_outputs),
    SEnum16('bounds_mode',
        'clip',
        'clip_and_normalize',
        'scale_to_fit',
        ),
    QStruct('bounds', INCLUDE=from_to),

    Pad(6),
    dyn_senum16('turn_off_with', DYN_NAME_PATH="..[DYN_I].usage"),
    Float('scale_by'),

    Pad(268),
    ascii_str32('usage'),

    SIZE=360
    )

permutation = Struct('permutation',
    Float('weight'),
    QStruct('color_lower_bound', INCLUDE=rgb_float),
    QStruct('color_upper_bound', INCLUDE=rgb_float),

    SIZE=28
    )

change_color = Struct('change_color',
    SEnum16('darken_by', *function_inputs_outputs),
    SEnum16('scale_by', *function_inputs_outputs),
    Bool32('flags',
        'blend_in_hsv',
        'more_colors',
        ),
    QStruct('color_lower_bound', INCLUDE=rgb_float),
    QStruct('color_upper_bound', INCLUDE=rgb_float),
    reflexive("permutations", permutation, 8),

    SIZE=44
    )

obje_attrs = Struct('obje_attrs',
    FlSEnum16("object_type",
        *((object_types[i], i - 1) for i in
          range(len(object_types))),
        VISIBLE=False, DEFAULT=-1
        ),
    Bool16('flags',
        'does_not_cast_shadow',
        'transparent_self_occlusion',
        'brighter_than_it_should_be',
        'not_a_pathfinding_obstacle',
        {NAME: 'xbox_unknown_bit_8', VALUE: 1<<8, VISIBLE: False},
        {NAME: 'xbox_unknown_bit_11', VALUE: 1<<11, VISIBLE: False},
        ),
    float_wu('bounding_radius'),
    QStruct('bounding_offset', INCLUDE=xyz_float),
    QStruct('origin_offset', INCLUDE=xyz_float),
    float_zero_to_inf('acceleration_scale', UNIT_SCALE=per_sec_unit_scale),

    Pad(4),
    dependency('model', valid_models),
    dependency('animation_graph', "antr"),

    Pad(40),
    dependency('collision_model', "coll"),
    dependency('physics', "phys"),
    dependency('modifier_shader', valid_shaders),
    dependency('creation_effect', "effe"),
    Pad(84),
    float_wu('render_bounding_radius'),

    #Export to functions
    SEnum16('A_in', *object_export_to),
    SEnum16('B_in', *object_export_to),
    SEnum16('C_in', *object_export_to),
    SEnum16('D_in', *object_export_to),

    Pad(44),
    SInt16('hud_text_message_index'),
    SInt16('forced_shader_permutation_index'),
    reflexive("attachments", attachment, 8, DYN_NAME_PATH='.type.filepath'),
    reflexive("widgets", widget, 4, DYN_NAME_PATH='.reference.filepath'),
    reflexive("functions", function, 4, DYN_NAME_PATH='.usage'),
    reflexive("change_colors", change_color, 4,
        'A', 'B', 'C', 'D'),
    reflexive("predicted_resources", predicted_resource, 1024, VISIBLE=False),

    SIZE=380
    )

obje_body = Struct('tagdata',
    obje_attrs,
    SIZE=380
    )

obje_def = TagDef("obje",
    blam_header('obje'),
    obje_body,

    ext=".object", endian=">", tag_cls=ObjeTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: -DeToX-
# 	Mapped plugin structure a new.
# revision: 2		author: Lord Zedd
# 	Standardizing.
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


cfxs_body = Struct("tagdata",
    Bool16("flags_0",
        "disable_brightness",
        ("disable_overexposure", 1 << 2),
        ),
    SInt16("unknown_0", VISIBLE=False),
    Float("overexposure_amount"),
    Float("overexposure_unknown_0"),
    Float("overexposure_unknown_1"),
    Float("brightness_amount"),
    Float("brightness_unknown_0"),
    Float("brightness_unknown_1"),
    Float("brightness_unknown_2"),
    Bool16("flags_1",
        "disable",
        ),
    SInt16("unknown_1", VISIBLE=False),
    Float("unknown_2"),
    Bool16("flags_2",
        "disable",
        ),
    SInt16("unknown_3", VISIBLE=False),
    Float("unknown_4"),
    Bool16("flags_3",
        "disable",
        ),
    SInt16("unknown_5", VISIBLE=False),
    Float("base_0"),
    Float("min_0"),
    Float("max_0"),
    Bool16("flags_4",
        "disable",
        ),
    SInt16("unknown_6", VISIBLE=False),
    Float("base_1"),
    Float("min_1"),
    Float("max_1"),
    Bool16("flags_5",
        "disable",
        ),
    SInt16("unknown_7", VISIBLE=False),
    Float("base_2"),
    Float("min_2"),
    Float("max_2"),
    Bool16("flags_6",
        "disable",
        ),
    SInt16("unknown_8", VISIBLE=False),
    color_rgb_float("color_0"),
    Bool16("flags_7",
        "disable",
        ),
    SInt16("unknown_9", VISIBLE=False),
    color_rgb_float("color_1"),
    Bool16("flags_8",
        "disable",
        ),
    SInt16("unknown_10", VISIBLE=False),
    color_rgb_float("color_2"),
    Bool16("flags_9",
        "disable",
        ),
    SInt16("unknown_11", VISIBLE=False),
    Float("unknown_12"),
    Float("unknown_13"),
    Float("unknown_14"),
    Bool16("flags_10",
        "disable",
        ),
    SInt16("unknown_15", VISIBLE=False),
    Float("unknown_16"),
    Float("unknown_17"),
    Float("unknown_18"),
    Bool16("flags_11",
        "disable",
        ),
    SInt16("unknown_19", VISIBLE=False),
    Float("unknown_20"),
    Float("unknown_21"),
    Float("unknown_22"),
    SInt32("unknown_23"),
    Bool16("flags_12",
        "disable",
        ),
    SInt16("unknown_24", VISIBLE=False),
    Float("base_3"),
    Float("min_3"),
    Float("max_3"),
    Bool16("flags_13",
        "disable",
        ),
    SInt16("unknown_25", VISIBLE=False),
    Float("base_4"),
    Float("min_4"),
    Float("max_4"),
    ENDIAN=">", SIZE=228
    )


def get():
    return cfxs_def

cfxs_def = TagDef("cfxs",
    h3_blam_header('cfxs'),
    cfxs_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["cfxs"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from traceback import format_exc

from reclaimer.sounds import constants, ogg, util, adpcm


class BlamSoundSamples:
    _sample_data = b''
    _sample_count = 0
    _compression = constants.COMPRESSION_PCM_16_LE
    _sample_rate = constants.SAMPLE_RATE_22K
    _encoding = constants.ENCODING_MONO
    _mouth_data = b''
    def __init__(self, sample_data, sample_count, compression,
                 sample_rate, encoding, mouth_data=b''):
        self._sample_data = sample_data
        self._sample_count = sample_count
        self._compression = compression
        self._sample_rate = sample_rate
        self._encoding = encoding
        self._mouth_data = mouth_data

    @property
    def sample_data(self):
        return self._sample_data
    @property
    def mouth_data(self):
        return self._mouth_data
    @property
    def sample_count(self):
        return self._sample_count
    @property
    def compression(self):
        return self._compression
    @property
    def sample_rate(self):
        return self._sample_rate
    @property
    def encoding(self):
        return self._encoding
    @property
    def is_compressed(self):
        return self.compression not in constants.PCM_FORMATS

    def compress(self, target_compression, target_sample_rate=None,
                 target_encoding=None, **compressor_kwargs):
        if target_sample_rate is None:
            target_sample_rate = self.sample_rate

        if target_encoding is None:
            target_encoding = self.encoding

        if (target_compression == self.compression and
            target_sample_rate == self.sample_rate and
            target_encoding == self.encoding):
            # compressing to same settings. nothing to do.
            return

        if (target_compression == constants.COMPRESSION_OGG and
              not constants.OGG_VORBIS_AVAILABLE):
            raise NotImplementedError(
                "Ogg encoder not available. Cannot compress.")
        elif (target_compression not in constants.PCM_FORMATS and
              target_compression != constants.COMPRESSION_XBOX_ADPCM and
              target_compression != constants.COMPRESSION_IMA_ADPCM and
              target_compression != constants.COMPRESSION_OGG):
            raise ValueError('Unknown compression type "%s"' %
                             target_compression)
        elif target_encoding not in (constants.ENCODING_MONO,
                                     constants.ENCODING_STEREO):
            raise ValueError("Compression encoding must be mono or stereo.")
        elif target_sample_rate <= 0:
            raise ValueError("Sample rate must be greater than zero.")

        sample_data = self.sample_data
        compression = self.compression
        if self.is_compressed:
            # decompress samples so we can recompress(ugh, whatever, i don't care)
            compression = (
                constants.ADPCM_DECOMPRESSED_FORMAT if
                compression in (constants.COMPRESSION_XBOX_ADPCM,
                                constants.COMPRESSION_IMA_ADPCM)
                else constants.HIGHEST_FIDELITY_FORMAT)
            sample_data = self.get_decompressed(
                compression, self.sample_rate, self.encoding)

        # make sure the sample data is the correct sample rate
        # and encoding before we try to compress it
        if (self.encoding != target_encoding or
            self.sample_rate != target_sample_rate):
            sample_data = util.convert_pcm_to_pcm(
                sample_data,
                self.compression, self.compression,
                self.encoding, target_encoding,
                self.sample_rate, target_sample_rate)

        if target_compression in (constants.COMPRESSION_XBOX_ADPCM,
                                  constants.COMPRESSION_IMA_ADPCM):
            # compress to adpcm
            if constants.sample_widths[compression] != 2:
                # adpcm encoder only accepts 16bit pcm as input
                sample_data = util.convert_pcm_to_pcm(
                    sample_data, compression,
                    constants.ADPCM_DECOMPRESSED_FORMAT)
                compression = constants.ADPCM_DECOMPRESSED_FORMAT

            adpcm_kwargs = compressor_kwargs.get("adpcm_kwargs", {})

            sample_data = adpcm.encode_adpcm_samples(
                sample_data, constants.channel_counts[target_encoding],
                util.is_big_endian_pcm(compression), **adpcm_kwargs)
        elif target_compression == constants.COMPRESSION_OGG:
            # compress to ogg vorbis
            # TODO: Finish this
            ogg_kwargs = compressor_kwargs.get("ogg_kwargs", {})

            raise NotImplementedError("Whoops, ogg is not implemented.")
        elif target_compression != self.compression:
            # convert to a different pcm format
            sample_data = util.convert_pcm_to_pcm(
                sample_data, self.compression, target_compression)

        self._sample_data = sample_data
        self._compression = target_compression
        self._encoding = target_encoding
        self._sample_rate = target_sample_rate

    def get_decompressed(self, target_compression, target_sample_rate=None,
                         target_encoding=None):
        if target_encoding is None:
            target_encoding = self.encoding

        if target_sample_rate is None:
            target_sample_rate = self.sample_rate

        assert target_compression in constants.PCM_FORMATS
        assert target_encoding in constants.channel_counts
        assert target_sample_rate > 0

        curr_compression = self.compression
        if curr_compression in (constants.COMPRESSION_XBOX_ADPCM,
                                constants.COMPRESSION_IMA_ADPCM):
            # decompress adpcm to 16bit pcm
            sample_data = adpcm.decode_adpcm_samples(
                self.sample_data, constants.channel_counts[self.encoding],
                util.is_big_endian_pcm(target_compression))
            curr_compression = constants.ADPCM_DECOMPRESSED_FORMAT
        elif not self.is_compressed:
            # samples are decompressed. use as-is
            sample_data = self.sample_data
        elif curr_compression == constants.COMPRESSION_OGG:
            if not constants.OGG_VORBIS_AVAILABLE:
                raise NotImplementedError(
                    "Ogg decoder not available. Cannot decompress.")
            # TODO: Finish this
        elif curr_compression == constants.COMPRESSION_WMA:
            if not constants.WMA_AVAILABLE:
                raise NotImplementedError(
                    "Wma decoder not available. Cannot decompress.")
            # TODO: Finish this
        else:
            raise ValueError("Unknown compression format.")

        if (curr_compression != target_compression or
            self.encoding != target_encoding or
            self.sample_rate != target_sample_rate):
            sample_data = util.convert_pcm_to_pcm(
                sample_data, curr_compression, target_compression,
                self.encoding, target_encoding,
                self.sample_rate, target_sample_rate)

        return sample_data

    def generate_mouth_data(self):
        # decompress to constants.DEFAULT_UNCOMPRESSED_FORMAT
        # as it is guaranteed to be a decompressed PCM format,
        # is guaranteed to be the correct system endianness,
        # and is high enough fidelity to generate mouth_data.
        sample_data = self.get_decompressed(
            constants.DEFAULT_UNCOMPRESSED_FORMAT,
            self.sample_rate, self.encoding)

        self._mouth_data = util.generate_mouth_data(
            sample_data, constants.DEFAULT_UNCOMPRESSED_FORMAT,
            self.sample_rate, self.encoding)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: -DeToX-
# 	Labelled Raw Identifier
# revision: 3		author: Xerax
#
# revision: 4		author: Lord Zedd
# 	Standardized
# revision: 5		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


snd__body = Struct("tagdata",
    Bool16("flags",
        "fit_to_adpcm_blocksize",
        "split_long_sound_into_permutations",
        "always spatialize",
        "never obstruct",
        {NAME: "INTERNAL DONT TOUCH", EDITABLE:False},
        "use huge sound transmission"
        "link count to owner unit",
        "pitch range is language",
        "dont use sound class speaker flags",
        "dont use lipsync data",
        ),
    SEnum8("sound_class", *snd__sound_class),
    UEnum8("compression",
        "xma",
        ),
    SInt16("platform_codec_index"),
    SInt16("pitch_range_index"),
    SInt16("language_b_index"),
    SInt16("unknown_1", VISIBLE=False),
    SInt16("playback_parameter_index"),
    SInt16("scale_index"),
    SInt8("promotion_index"),
    SInt8("custom_playback_index"),
    SInt16("extra_info_index"),
    SInt32("maximum_play_time"),
    zone_asset("asset_data"),
    ENDIAN=">", SIZE=32
    )


def get():
    return snd__def

snd__def = TagDef("snd!",
    h3_blam_header('snd!'),
    snd__body,

    ext=".%s" % h3_tag_class_fcc_to_ext["snd!"], endian=">", tag_cls=H3Tag
    )
